from autolens import text_util
from autolens.dimensions import (
    DimensionsProfile,
    Length,
    Luminosity,
    Mass,
    MassOverLuminosity,
    MassOverLength2,
    MassOverLength3,
    Position,
    convert_units_to_input_units,
)
from autolens.array.mask import Mask, load_mask_from_fits, output_mask_to_fits
from autolens.array.mapping import Mapping
from autolens.array import mapping
from autolens.array import grids
from autolens.array.grids import (
    Grid,
    BinnedGrid,
    PixelizationGrid,
    SparseToGrid,
    Interpolator,
)
from autolens.array.scaled_array import (
    ArrayGeometry,
    RectangularArrayGeometry,
    Array,
    ScaledArray,
    ScaledSquarePixelArray,
    ScaledRectangularPixelArray,
)
from autolens.array.convolution import Convolver
from autolens.array.fourier_transform import Transformer
from autolens.array.util import array_util, binning_util, grid_util, mask_util
from autolens.array.mapping_util import (
    array_mapping_util,
    grid_mapping_util,
    mask_mapping_util,
    sparse_mapping_util,
)
from autolens.data import abstract_data
from autolens.data import uv_plane
from autolens.data.abstract_data import (
    AbstractData,
    AbstractNoiseMap,
    ExposureTimeMap,
    load_image,
    load_exposure_time_map,
    load_positions,
    output_positions,
)
from autolens.data.imaging import (
    ImagingData,
    PSF,
    NoiseMap,
    PoissonNoiseMap,
    SimulatedImagingData,
    generate_poisson_noise,
    load_imaging_data_from_fits,
    load_noise_map,
    load_psf,
    output_imaging_data_to_fits,
)
from autolens.data.uv_plane import (
    UVPlaneData,
    PrimaryBeam,
    SimulatedUVPlaneData,
    load_uv_plane_data_from_fits,
    output_uv_plane_data_to_fits,
    gaussian_noise_map_from_shape_and_sigma,
)
from autolens.data.plotters import data_plotters, imaging_plotters, uv_plane_plotters
from autolens.lens import ray_tracing
from autolens.lens.lens_data import AbstractLensData, LensImagingData, LensUVPlaneData
from autolens.lens.lens_fit import (
    ImagingFit,
    LensImagingFit,
    UVPlaneFit,
    LensUVPlaneFit,
    LensPositionFit,
)
from autolens.lens.plane import Plane, PlanePositions, PlaneImage
from autolens.lens.plotters import (
    lens_imaging_fit_plotters,
    lens_plotter_util,
    plane_plotters,
    ray_tracing_plotters,
)
from autolens.lens.ray_tracing import Tracer
from autolens.lens.util import lens_util
from autolens.model import cosmology_util
from autolens.model.galaxy.galaxy import Galaxy, HyperGalaxy, Redshift
from autolens.model.galaxy.galaxy_data import GalaxyData, GalaxyFitData
from autolens.model.galaxy.galaxy_fit import GalaxyFit
from autolens.model.galaxy.galaxy_model import GalaxyModel
from autolens.model.galaxy.plotters import galaxy_fit_plotters, galaxy_plotters
from autolens.model.hyper.hyper_data import HyperImageSky, HyperBackgroundNoise
from autolens.model.inversion.inversions import Inversion
from autolens.model.inversion.mappers import Mapper, RectangularMapper, VoronoiMapper
from autolens.model.inversion import pixelizations, regularization
from autolens.model.inversion.plotters import inversion_plotters, mapper_plotters
from autolens.model.inversion.util import (
    inversion_util,
    mapper_util,
    pixelization_util,
    regularization_util,
)
from autolens.model.profiles import (
    geometry_profiles,
    light_profiles,
    mass_profiles,
    light_and_mass_profiles,
)
from autolens.model.profiles.plotters import profile_plotters
from autolens.pipeline import phase_tagging, pipeline_tagging
from autolens.pipeline.phase import phase
from autolens.pipeline.phase.phase import AbstractPhase
from autolens.pipeline.phase.phase_extensions import CombinedHyperPhase
from autolens.pipeline.phase.phase_extensions import HyperGalaxyPhase
from autolens.pipeline.phase.phase_extensions.hyper_galaxy_phase import HyperGalaxyPhase
from autolens.pipeline.phase.phase_extensions.hyper_phase import HyperPhase
from autolens.pipeline.phase.phase_extensions.inversion_phase import (
    InversionBackgroundBothPhase,
    InversionBackgroundNoisePhase,
    InversionBackgroundSkyPhase,
    InversionPhase,
    VariableFixingHyperPhase,
)
from autolens.pipeline.phase.phase import AbstractPhase
from autolens.pipeline.phase.phase_data import PhaseData
from autolens.pipeline.phase.phase_imaging import PhaseImaging
from autolens.pipeline.phase.phase_galaxy import PhaseGalaxy
from autolens.pipeline.pipeline import (
    PipelineSettings,
    PipelineSettingsHyper,
    PipelineImaging,
    PipelinePositions,
)
from autolens.pipeline.plotters import hyper_plotters, phase_plotters
from autolens.plotters import array_plotters, grid_plotters, plotter_util

__version__ = '0.31.7'



import numpy as np
from astropy import cosmology as cosmo

import autofit as af
from autolens.array.util import binning_util
from autolens.lens import lens_data as ld
from autolens.lens import lens_fit
from autolens.model.galaxy import galaxy as g
from autolens.pipeline import phase_tagging
from autolens.pipeline.phase import phase_extensions
from autolens.pipeline.phase.phase_data import PhaseData
from autolens.plotters import visualizer


class PhaseImaging(PhaseData):
    galaxies = af.PhaseProperty("galaxies")
    hyper_image_sky = af.PhaseProperty("hyper_image_sky")
    hyper_background_noise = af.PhaseProperty("hyper_background_noise")

    def __init__(
        self,
        phase_name,
        phase_folders=tuple(),
        galaxies=None,
        hyper_image_sky=None,
        hyper_background_noise=None,
        optimizer_class=af.MultiNest,
        cosmology=cosmo.Planck15,
        sub_size=2,
        signal_to_noise_limit=None,
        bin_up_factor=None,
        psf_shape=None,
        positions_threshold=None,
        mask_function=None,
        inner_mask_radii=None,
        pixel_scale_interpolation_grid=None,
        pixel_scale_binned_cluster_grid=None,
        inversion_uses_border=True,
        inversion_pixel_limit=None,
        auto_link_priors=False,
    ):

        """

        A phase in an lens pipeline. Uses the set non_linear optimizer to try to fit models and hyper_galaxies
        passed to it.

        Parameters
        ----------
        optimizer_class: class
            The class of a non_linear optimizer
        sub_size: int
            The side length of the subgrid
        pixel_scale_binned_cluster_grid : float or None
            If *True*, the hyper_galaxies image used to generate the cluster'grids weight map will be binned \
            up to this higher pixel scale to speed up the KMeans clustering algorithm.
        """

        phase_tag = phase_tagging.phase_tag_from_phase_settings(
            sub_size=sub_size,
            signal_to_noise_limit=signal_to_noise_limit,
            bin_up_factor=bin_up_factor,
            psf_shape=psf_shape,
            positions_threshold=positions_threshold,
            inner_mask_radii=inner_mask_radii,
            pixel_scale_interpolation_grid=pixel_scale_interpolation_grid,
            pixel_scale_binned_cluster_grid=pixel_scale_binned_cluster_grid,
        )

        super(PhaseImaging, self).__init__(
            phase_name=phase_name,
            phase_tag=phase_tag,
            phase_folders=phase_folders,
            galaxies=galaxies,
            sub_size=sub_size,
            signal_to_noise_limit=signal_to_noise_limit,
            positions_threshold=positions_threshold,
            mask_function=mask_function,
            inner_mask_radii=inner_mask_radii,
            pixel_scale_interpolation_grid=pixel_scale_interpolation_grid,
            pixel_scale_binned_cluster_grid=pixel_scale_binned_cluster_grid,
            inversion_uses_border=inversion_uses_border,
            inversion_pixel_limit=inversion_pixel_limit,
            optimizer_class=optimizer_class,
            cosmology=cosmology,
            auto_link_priors=auto_link_priors,
        )

        self.bin_up_factor = bin_up_factor
        self.psf_shape = psf_shape

        self.hyper_image_sky = hyper_image_sky
        self.hyper_background_noise = hyper_background_noise

        self.hyper_noise_map_max = af.conf.instance.general.get(
            "hyper", "hyper_noise_map_max", float
        )

        self.is_hyper_phase = False

    # noinspection PyMethodMayBeStatic,PyUnusedLocal
    def modify_image(self, image, results):
        """
        Customize an lens_data. e.g. removing lens light.

        Parameters
        ----------
        image: scaled_array.ScaledSquarePixelArray
            An lens_data that has been masked
        results: autofit.tools.pipeline.ResultsCollection
            The result of the previous lens

        Returns
        -------
        lens_data: scaled_array.ScaledSquarePixelArray
            The modified image (not changed by default)
        """
        return image

    def make_analysis(self, data, results=None, mask=None, positions=None):
        """
        Create an lens object. Also calls the prior passing and lens_data modifying functions to allow child
        classes to change the behaviour of the phase.

        Parameters
        ----------
        positions
        mask: Mask
            The default masks passed in by the pipeline
        data: im.Imaging
            An lens_data that has been masked
        results: autofit.tools.pipeline.ResultsCollection
            The result from the previous phase

        Returns
        -------
        lens : Analysis
            An lens object that the non-linear optimizer calls to determine the fit of a set of values
        """

        mask = self.setup_phase_mask(data=data, mask=mask)

        self.check_positions(positions=positions)

        pixel_scale_binned_grid = self.pixel_scale_binned_grid_from_mask(mask=mask)

        preload_pixelization_grids_of_planes = self.preload_pixelization_grids_of_planes_from_results(
            results=results
        )

        lens_imaging_data = ld.LensImagingData(
            imaging_data=data,
            mask=mask,
            trimmed_psf_shape=self.psf_shape,
            positions=positions,
            positions_threshold=self.positions_threshold,
            pixel_scale_interpolation_grid=self.pixel_scale_interpolation_grid,
            pixel_scale_binned_grid=pixel_scale_binned_grid,
            hyper_noise_map_max=self.hyper_noise_map_max,
            inversion_pixel_limit=self.inversion_pixel_limit,
            inversion_uses_border=self.inversion_uses_border,
            preload_pixelization_grids_of_planes=preload_pixelization_grids_of_planes,
        )

        modified_image = self.modify_image(
            image=lens_imaging_data.image(return_in_2d=True, return_masked=False),
            results=results,
        )

        lens_imaging_data = lens_imaging_data.new_lens_imaging_data_with_modified_image(
            modified_image=modified_image
        )

        if self.signal_to_noise_limit is not None:
            lens_imaging_data = lens_imaging_data.new_lens_imaging_data_with_signal_to_noise_limit(
                signal_to_noise_limit=self.signal_to_noise_limit
            )

        if self.bin_up_factor is not None:
            lens_imaging_data = lens_imaging_data.new_lens_imaging_data_with_binned_up_imaging_data_and_mask(
                bin_up_factor=self.bin_up_factor
            )

        self.output_phase_info()

        analysis = self.Analysis(
            lens_imaging_data=lens_imaging_data,
            cosmology=self.cosmology,
            image_path=self.optimizer.image_path,
            results=results,
        )

        return analysis

    def output_phase_info(self):

        file_phase_info = "{}/{}".format(self.optimizer.phase_output_path, "phase.info")

        with open(file_phase_info, "w") as phase_info:
            phase_info.write("Optimizer = {} \n".format(type(self.optimizer).__name__))
            phase_info.write("Sub-grid size = {} \n".format(self.sub_size))
            phase_info.write("PSF shape = {} \n".format(self.psf_shape))
            phase_info.write(
                "Positions Threshold = {} \n".format(self.positions_threshold)
            )
            phase_info.write("Cosmology = {} \n".format(self.cosmology))
            phase_info.write("Auto Link Priors = {} \n".format(self.auto_link_priors))

            phase_info.close()

    def extend_with_multiple_hyper_phases(
        self,
        hyper_galaxy=False,
        inversion=False,
        include_background_sky=False,
        include_background_noise=False,
    ):

        if hyper_galaxy:
            if not include_background_sky and not include_background_noise:
                phase_hyper_galaxy = (
                    phase_extensions.hyper_galaxy_phase.HyperGalaxyPhase
                )
            elif include_background_sky and not include_background_noise:
                phase_hyper_galaxy = (
                    phase_extensions.hyper_galaxy_phase.HyperGalaxyBackgroundSkyPhase
                )
            elif not include_background_sky and include_background_noise:
                phase_hyper_galaxy = (
                    phase_extensions.hyper_galaxy_phase.HyperGalaxyBackgroundNoisePhase
                )
            else:
                phase_hyper_galaxy = (
                    phase_extensions.hyper_galaxy_phase.HyperGalaxyBackgroundBothPhase
                )
        else:
            phase_hyper_galaxy = None

        if inversion:
            if not include_background_sky and not include_background_noise:
                phase_inversion = phase_extensions.InversionPhase
            elif include_background_sky and not include_background_noise:
                phase_inversion = phase_extensions.InversionBackgroundSkyPhase
            elif not include_background_sky and include_background_noise:
                phase_inversion = phase_extensions.InversionBackgroundNoisePhase
            else:
                phase_inversion = phase_extensions.InversionBackgroundBothPhase
        else:
            phase_inversion = None

        hyper_phase_classes = tuple(filter(None, (phase_hyper_galaxy, phase_inversion)))

        if len(hyper_phase_classes) == 0:
            return self
        else:
            return phase_extensions.CombinedHyperPhase(
                phase=self, hyper_phase_classes=hyper_phase_classes
            )

    # noinspection PyAbstractClass
    class Analysis(PhaseData.Analysis):
        def __init__(self, lens_imaging_data, cosmology, image_path=None, results=None):
            super(PhaseImaging.Analysis, self).__init__(
                cosmology=cosmology, results=results
            )
            self.visualizer = visualizer.PhaseImagingVisualizer(
                lens_imaging_data, image_path
            )

            self.lens_imaging_data = lens_imaging_data

            self.visualizer.plot_hyper_images(self.last_results)

            if self.last_results is not None:
                self.hyper_galaxy_image_1d_path_dict = (
                    self.last_results.hyper_galaxy_image_1d_path_dict
                )

                self.hyper_model_image_1d = self.last_results.hyper_model_image_1d

                self.binned_hyper_galaxy_image_1d_path_dict = self.last_results.binned_hyper_galaxy_image_1d_path_dict(
                    binned_grid=lens_imaging_data.grid.binned
                )

                self.visualizer.plot_hyper_images(self.last_results)

        @property
        def lens_data(self):
            return self.lens_imaging_data

        def fit(self, instance):
            """
            Determine the fit of a lens galaxy and source galaxy to the lens_data in this lens.

            Parameters
            ----------
            instance
                A model instance with attributes

            Returns
            -------
            fit : Fit
                A fractional value indicating how well this model fit and the model lens_data itself
            """

            self.associate_images(instance=instance)
            tracer = self.tracer_for_instance(instance=instance)

            self.check_positions_trace_within_threshold_via_tracer(tracer=tracer)
            self.check_inversion_pixels_are_below_limit_via_tracer(tracer=tracer)

            hyper_image_sky = self.hyper_image_sky_for_instance(instance=instance)

            hyper_background_noise = self.hyper_background_noise_for_instance(
                instance=instance
            )

            fit = self.lens_imaging_fit_for_tracer(
                tracer=tracer,
                hyper_image_sky=hyper_image_sky,
                hyper_background_noise=hyper_background_noise,
            )

            return fit.figure_of_merit

        def associate_images(self, instance: af.ModelInstance) -> af.ModelInstance:
            """
            Takes images from the last result, if there is one, and associates them with galaxies in this phase
            where full-path galaxy names match.

            If the galaxy collection has a different name then an association is not made.

            e.g.
            galaxies.lens will match with:
                galaxies.lens
            but not with:
                galaxies.lens
                galaxies.source

            Parameters
            ----------
            instance
                A model instance with 0 or more galaxies in its tree

            Returns
            -------
            instance
               The input instance with images associated with galaxies where possible.
            """
            if hasattr(self, "hyper_galaxy_image_1d_path_dict"):
                for galaxy_path, galaxy in instance.path_instance_tuples_for_class(
                    g.Galaxy
                ):
                    if galaxy_path in self.hyper_galaxy_image_1d_path_dict:
                        galaxy.hyper_model_image_1d = self.hyper_model_image_1d
                        galaxy.hyper_galaxy_image_1d = self.hyper_galaxy_image_1d_path_dict[
                            galaxy_path
                        ]
                        if (
                            hasattr(self, "binned_hyper_galaxy_image_1d_path_dict")
                            and self.binned_hyper_galaxy_image_1d_path_dict is not None
                        ):
                            galaxy.binned_hyper_galaxy_image_1d = self.binned_hyper_galaxy_image_1d_path_dict[
                                galaxy_path
                            ]
            return instance

        def hyper_image_sky_for_instance(self, instance):

            if hasattr(instance, "hyper_image_sky"):
                return instance.hyper_image_sky
            else:
                return None

        def hyper_background_noise_for_instance(self, instance):

            if hasattr(instance, "hyper_background_noise"):
                return instance.hyper_background_noise
            else:
                return None

        def lens_imaging_fit_for_tracer(
            self, tracer, hyper_image_sky, hyper_background_noise
        ):

            return lens_fit.LensImagingFit.from_lens_data_and_tracer(
                lens_data=self.lens_imaging_data,
                tracer=tracer,
                hyper_image_sky=hyper_image_sky,
                hyper_background_noise=hyper_background_noise,
            )

        def visualize(self, instance, during_analysis):
            instance = self.associate_images(instance=instance)
            tracer = self.tracer_for_instance(instance=instance)
            hyper_image_sky = self.hyper_image_sky_for_instance(instance=instance)
            hyper_background_noise = self.hyper_background_noise_for_instance(
                instance=instance
            )

            fit = self.lens_imaging_fit_for_tracer(
                tracer=tracer,
                hyper_image_sky=hyper_image_sky,
                hyper_background_noise=hyper_background_noise,
            )
            self.visualizer.plot_ray_tracing(fit.tracer, during_analysis)
            self.visualizer.plot_lens_imaging(fit, during_analysis)

    class Result(PhaseData.Result):
        @property
        def most_likely_fit(self):

            hyper_image_sky = self.analysis.hyper_image_sky_for_instance(
                instance=self.constant
            )

            hyper_background_noise = self.analysis.hyper_background_noise_for_instance(
                instance=self.constant
            )

            return self.analysis.lens_imaging_fit_for_tracer(
                tracer=self.most_likely_tracer,
                hyper_image_sky=hyper_image_sky,
                hyper_background_noise=hyper_background_noise,
            )

        @property
        def unmasked_model_image(self):
            return self.most_likely_fit.unmasked_blurred_profile_image

        @property
        def unmasked_model_image_of_planes(self):
            return self.most_likely_fit.unmasked_blurred_profile_image_of_planes

        @property
        def unmasked_model_image_of_planes_and_galaxies(self):
            fit = self.most_likely_fit
            return fit.unmasked_blurred_profile_image_of_planes_and_galaxies

        def image_2d_for_galaxy(self, galaxy: g.Galaxy) -> np.ndarray:
            """
            Parameters
            ----------
            galaxy
                A galaxy used in this phase

            Returns
            -------
            ndarray or None
                A numpy array giving the model image of that galaxy
            """
            return self.most_likely_fit.galaxy_model_image_2d_dict[galaxy]

        @property
        def image_galaxy_1d_dict(self) -> {str: g.Galaxy}:
            """
            A dictionary associating galaxy names with model images of those galaxies
            """

            image_1d_dict = {}

            for galaxy, galaxy_image_2d in self.image_galaxy_2d_dict.items():
                image_1d_dict[galaxy] = self.mask.mapping.array_1d_from_array_2d(
                    array_2d=galaxy_image_2d
                )

            return image_1d_dict

        @property
        def image_galaxy_2d_dict(self) -> {str: g.Galaxy}:
            """
            A dictionary associating galaxy names with model images of those galaxies
            """
            return {
                galaxy_path: self.image_2d_for_galaxy(galaxy)
                for galaxy_path, galaxy in self.path_galaxy_tuples
            }

        @property
        def hyper_galaxy_image_1d_path_dict(self):
            """
            A dictionary associating 1D hyper_galaxies galaxy images with their names.
            """

            hyper_minimum_percent = af.conf.instance.general.get(
                "hyper", "hyper_minimum_percent", float
            )

            hyper_galaxy_image_1d_path_dict = {}

            for path, galaxy in self.path_galaxy_tuples:

                galaxy_image_1d = self.image_galaxy_1d_dict[path]

                if not np.all(galaxy_image_1d == 0):
                    minimum_galaxy_value = hyper_minimum_percent * max(galaxy_image_1d)
                    galaxy_image_1d[
                        galaxy_image_1d < minimum_galaxy_value
                    ] = minimum_galaxy_value

                hyper_galaxy_image_1d_path_dict[path] = galaxy_image_1d

            return hyper_galaxy_image_1d_path_dict

        @property
        def hyper_galaxy_image_2d_path_dict(self):
            """
            A dictionary associating 2D hyper_galaxies galaxy images with their names.
            """

            hyper_galaxy_image_2d_path_dict = {}

            for path, galaxy in self.path_galaxy_tuples:
                hyper_galaxy_image_2d_path_dict[
                    path
                ] = self.mask.mapping.scaled_array_2d_from_array_1d(
                    array_1d=self.hyper_galaxy_image_1d_path_dict[path]
                )

            return hyper_galaxy_image_2d_path_dict

        def binned_image_1d_dict_from_binned_grid(self, binned_grid) -> {str: g.Galaxy}:
            """
            A dictionary associating 1D binned images with their names.
            """

            binned_image_1d_dict = {}

            for galaxy, galaxy_image_2d in self.image_galaxy_2d_dict.items():
                binned_image_2d = binning_util.binned_up_array_2d_using_mean_from_array_2d_and_bin_up_factor(
                    array_2d=galaxy_image_2d, bin_up_factor=binned_grid.bin_up_factor
                )

                binned_image_1d_dict[
                    galaxy
                ] = binned_grid.mask.mapping.array_1d_from_array_2d(
                    array_2d=binned_image_2d
                )

            return binned_image_1d_dict

        def binned_hyper_galaxy_image_1d_path_dict(self, binned_grid):
            """
            A dictionary associating 1D hyper_galaxies galaxy binned images with their names.
            """

            if binned_grid is not None:

                hyper_minimum_percent = af.conf.instance.general.get(
                    "hyper", "hyper_minimum_percent", float
                )

                binned_image_1d_galaxy_dict = self.binned_image_1d_dict_from_binned_grid(
                    binned_grid=binned_grid
                )

                binned_hyper_galaxy_image_path_dict = {}

                for path, galaxy in self.path_galaxy_tuples:
                    binned_galaxy_image_1d = binned_image_1d_galaxy_dict[path]

                    minimum_hyper_value = hyper_minimum_percent * max(
                        binned_galaxy_image_1d
                    )
                    binned_galaxy_image_1d[
                        binned_galaxy_image_1d < minimum_hyper_value
                    ] = minimum_hyper_value

                    binned_hyper_galaxy_image_path_dict[path] = binned_galaxy_image_1d

                return binned_hyper_galaxy_image_path_dict

        def binned_hyper_galaxy_image_2d_path_dict(self, binned_grid):
            """
            A dictionary associating "D hyper_galaxies galaxy images binned images with their names.
            """

            if binned_grid is not None:

                binned_hyper_galaxy_image_1d_path_dict = self.binned_hyper_galaxy_image_1d_path_dict(
                    binned_grid=binned_grid
                )

                binned_hyper_galaxy_image_2d_path_dict = {}

                for path, galaxy in self.path_galaxy_tuples:
                    binned_hyper_galaxy_image_2d_path_dict[
                        path
                    ] = binned_grid.mask.mapping.scaled_array_2d_from_array_1d(
                        array_1d=binned_hyper_galaxy_image_1d_path_dict[path]
                    )

                return binned_hyper_galaxy_image_2d_path_dict

        @property
        def hyper_model_image_1d(self):

            hyper_model_image_1d = np.zeros(self.mask.pixels_in_mask)

            for path, galaxy in self.path_galaxy_tuples:
                hyper_model_image_1d += self.hyper_galaxy_image_1d_path_dict[path]

            return hyper_model_image_1d



import autolens as al
import numpy as np
import pytest


class MockTracerPositions:
    def __init__(self, positions, noise=None):
        self.positions = positions
        self.noise = noise


class TestPositionFit:
    def test__x1_positions__mock_position_tracer__maximum_separation_is_correct(self):
        tracer = MockTracerPositions(positions=[np.array([[0.0, 0.0], [0.0, 1.0]])])
        fit = al.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.maximum_separations[0] == 1.0

        tracer = MockTracerPositions(positions=[np.array([[0.0, 0.0], [1.0, 1.0]])])
        fit = al.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.maximum_separations[0] == np.sqrt(2)

        tracer = MockTracerPositions(positions=[np.array([[0.0, 0.0], [1.0, 3.0]])])
        fit = al.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.maximum_separations[0] == np.sqrt(np.square(1.0) + np.square(3.0))

        tracer = MockTracerPositions(positions=[np.array([[-2.0, -4.0], [1.0, 3.0]])])
        fit = al.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.maximum_separations[0] == np.sqrt(np.square(3.0) + np.square(7.0))

        tracer = MockTracerPositions(positions=[np.array([[8.0, 4.0], [-9.0, -4.0]])])
        fit = al.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.maximum_separations[0] == np.sqrt(np.square(17.0) + np.square(8.0))

    def test_multiple_positions__mock_position_tracer__maximum_separation_is_correct(
        self
    ):
        tracer = MockTracerPositions(
            positions=[np.array([[0.0, 0.0], [0.0, 1.0], [0.0, 0.5]])]
        )
        fit = al.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.maximum_separations[0] == 1.0

        tracer = MockTracerPositions(
            positions=[np.array([[0.0, 0.0], [0.0, 0.0], [3.0, 3.0]])]
        )
        fit = al.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.maximum_separations[0] == np.sqrt(18)

        tracer = MockTracerPositions(
            positions=[np.array([[0.0, 0.0], [1.0, 1.0], [3.0, 3.0]])]
        )
        fit = al.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.maximum_separations[0] == np.sqrt(18)

        tracer = MockTracerPositions(
            positions=[
                np.array(
                    [
                        [-2.0, -4.0],
                        [1.0, 3.0],
                        [0.1, 0.1],
                        [-0.1, -0.1],
                        [0.3, 0.4],
                        [-0.6, 0.5],
                    ]
                )
            ]
        )
        fit = al.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.maximum_separations[0] == np.sqrt(np.square(3.0) + np.square(7.0))

        tracer = MockTracerPositions(
            positions=[np.array([[8.0, 4.0], [8.0, 4.0], [-9.0, -4.0]])]
        )
        fit = al.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.maximum_separations[0] == np.sqrt(np.square(17.0) + np.square(8.0))

    def test_multiple_sets_of_positions__multiple_sets_of_max_distances(self):
        tracer = MockTracerPositions(
            positions=[
                np.array([[0.0, 0.0], [0.0, 1.0], [0.0, 0.5]]),
                np.array([[0.0, 0.0], [0.0, 0.0], [3.0, 3.0]]),
                np.array([[0.0, 0.0], [1.0, 1.0], [3.0, 3.0]]),
            ]
        )

        fit = al.LensPositionFit(positions=tracer.positions, noise_map=1.0)

        assert fit.maximum_separations[0] == 1.0
        assert fit.maximum_separations[1] == np.sqrt(18)
        assert fit.maximum_separations[2] == np.sqrt(18)

    def test__likelihood__is_sum_of_separations_divided_by_noise(self):
        tracer = MockTracerPositions(
            positions=[
                np.array([[0.0, 0.0], [0.0, 1.0], [0.0, 0.5]]),
                np.array([[0.0, 0.0], [0.0, 0.0], [3.0, 3.0]]),
                np.array([[0.0, 0.0], [1.0, 1.0], [3.0, 3.0]]),
            ]
        )

        fit = al.LensPositionFit(positions=tracer.positions, noise_map=1.0)
        assert fit.chi_squared_map[0] == 1.0
        assert fit.chi_squared_map[1] == pytest.approx(18.0, 1e-4)
        assert fit.chi_squared_map[2] == pytest.approx(18.0, 1e-4)
        assert fit.figure_of_merit == pytest.approx(-0.5 * (1.0 + 18 + 18), 1e-4)

        fit = al.LensPositionFit(positions=tracer.positions, noise_map=2.0)
        assert fit.chi_squared_map[0] == (1.0 / 2.0) ** 2.0
        assert fit.chi_squared_map[1] == pytest.approx(
            (np.sqrt(18.0) / 2.0) ** 2.0, 1e-4
        )
        assert fit.chi_squared_map[2] == pytest.approx(
            (np.sqrt(18.0) / 2.0) ** 2.0, 1e-4
        )
        assert fit.figure_of_merit == pytest.approx(
            -0.5
            * (
                (1.0 / 2.0) ** 2.0
                + (np.sqrt(18.0) / 2.0) ** 2.0
                + (np.sqrt(18.0) / 2.0) ** 2.0
            ),
            1e-4,
        )

    def test__threshold__if_not_met_returns_ray_tracing_exception(self):
        tracer = MockTracerPositions(positions=[np.array([[0.0, 0.0], [0.0, 1.0]])])
        fit = al.LensPositionFit(positions=tracer.positions, noise_map=1.0)

        assert fit.maximum_separation_within_threshold(threshold=100.0)
        assert not fit.maximum_separation_within_threshold(threshold=0.1)



import autolens as al
import numpy as np
import pytest

import autofit as af
from test.unit.mock.model.mock_galaxy import MockGalaxy


class TestLikelihood:
    def test__1x1_image__light_profile_fits_data_perfectly__lh_is_noise(self):
        image = al.ScaledSquarePixelArray(array=np.ones((3, 3)), pixel_scale=1.0)

        noise_map = al.ScaledSquarePixelArray(array=np.ones((3, 3)), pixel_scale=1.0)

        galaxy_data = al.GalaxyData(image=image, noise_map=noise_map, pixel_scale=3.0)

        mask = al.Mask(
            array=np.array(
                [[True, True, True], [True, False, True], [True, True, True]]
            ),
            pixel_scale=1.0,
            sub_size=1,
        )
        g0 = MockGalaxy(value=1.0)

        galaxy_fit_data = al.GalaxyFitData(
            galaxy_data=galaxy_data, mask=mask, use_image=True
        )
        fit = al.GalaxyFit(galaxy_data=galaxy_fit_data, model_galaxies=[g0])
        assert fit.model_galaxies == [g0]
        assert fit.likelihood == -0.5 * np.log(2 * np.pi * 1.0)

        galaxy_fit_data = al.GalaxyFitData(
            galaxy_data=galaxy_data, mask=mask, use_convergence=True
        )
        fit = al.GalaxyFit(galaxy_data=galaxy_fit_data, model_galaxies=[g0])
        assert fit.model_galaxies == [g0]
        assert fit.likelihood == -0.5 * np.log(2 * np.pi * 1.0)

        galaxy_fit_data = al.GalaxyFitData(
            galaxy_data=galaxy_data, mask=mask, use_potential=True
        )
        fit = al.GalaxyFit(galaxy_data=galaxy_fit_data, model_galaxies=[g0])
        assert fit.model_galaxies == [g0]
        assert fit.likelihood == -0.5 * np.log(2 * np.pi * 1.0)

        galaxy_fit_data = al.GalaxyFitData(
            galaxy_data=galaxy_data, mask=mask, use_deflections_y=True
        )
        fit = al.GalaxyFit(galaxy_data=galaxy_fit_data, model_galaxies=[g0])
        assert fit.model_galaxies == [g0]
        assert fit.likelihood == -0.5 * np.log(2 * np.pi * 1.0)

        galaxy_fit_data = al.GalaxyFitData(
            galaxy_data=galaxy_data, mask=mask, use_deflections_x=True
        )
        fit = al.GalaxyFit(galaxy_data=galaxy_fit_data, model_galaxies=[g0])
        assert fit.model_galaxies == [g0]
        assert fit.likelihood == -0.5 * np.log(2 * np.pi * 1.0)

    def test__1x2_image__noise_not_1__alls_correct(self):
        image = al.ScaledSquarePixelArray(array=5.0 * np.ones((3, 4)), pixel_scale=1.0)
        image[1, 2] = 4.0

        noise_map = al.ScaledSquarePixelArray(
            array=2.0 * np.ones((3, 4)), pixel_scale=1.0
        )

        galaxy_data = al.GalaxyData(image=image, noise_map=noise_map, pixel_scale=3.0)

        mask = al.Mask(
            array=np.array(
                [
                    [True, True, True, True],
                    [True, False, False, True],
                    [True, True, True, True],
                ]
            ),
            pixel_scale=1.0,
            sub_size=1,
        )

        g0 = MockGalaxy(value=1.0, shape=2)

        galaxy_fit_data = al.GalaxyFitData(
            galaxy_data=galaxy_data, mask=mask, use_image=True
        )
        fit = al.GalaxyFit(galaxy_data=galaxy_fit_data, model_galaxies=[g0])

        assert fit.model_galaxies == [g0]
        assert fit.chi_squared == (25.0 / 4.0)
        assert fit.reduced_chi_squared == (25.0 / 4.0) / 2.0
        assert fit.likelihood == -0.5 * (
            (25.0 / 4.0) + 2.0 * np.log(2 * np.pi * 2.0 ** 2)
        )

        galaxy_fit_data = al.GalaxyFitData(
            galaxy_data=galaxy_data, mask=mask, use_convergence=True
        )
        fit = al.GalaxyFit(galaxy_data=galaxy_fit_data, model_galaxies=[g0])
        assert fit.model_galaxies == [g0]
        assert fit.chi_squared == (25.0 / 4.0)
        assert fit.reduced_chi_squared == (25.0 / 4.0) / 2.0
        assert fit.likelihood == -0.5 * (
            (25.0 / 4.0) + 2.0 * np.log(2 * np.pi * 2.0 ** 2)
        )

        galaxy_fit_data = al.GalaxyFitData(
            galaxy_data=galaxy_data, mask=mask, use_potential=True
        )
        fit = al.GalaxyFit(galaxy_data=galaxy_fit_data, model_galaxies=[g0])
        assert fit.model_galaxies == [g0]
        assert fit.chi_squared == (25.0 / 4.0)
        assert fit.reduced_chi_squared == (25.0 / 4.0) / 2.0
        assert fit.likelihood == -0.5 * (
            (25.0 / 4.0) + 2.0 * np.log(2 * np.pi * 2.0 ** 2)
        )

        galaxy_fit_data = al.GalaxyFitData(
            galaxy_data=galaxy_data, mask=mask, use_deflections_y=True
        )
        fit = al.GalaxyFit(galaxy_data=galaxy_fit_data, model_galaxies=[g0])
        assert fit.chi_squared == (25.0 / 4.0)
        assert fit.reduced_chi_squared == (25.0 / 4.0) / 2.0
        assert fit.likelihood == -0.5 * (
            (25.0 / 4.0) + 2.0 * np.log(2 * np.pi * 2.0 ** 2)
        )

        galaxy_fit_data = al.GalaxyFitData(
            galaxy_data=galaxy_data, mask=mask, use_deflections_x=True
        )
        fit = al.GalaxyFit(galaxy_data=galaxy_fit_data, model_galaxies=[g0])
        assert fit.chi_squared == (25.0 / 4.0)
        assert fit.reduced_chi_squared == (25.0 / 4.0) / 2.0
        assert fit.likelihood == -0.5 * (
            (25.0 / 4.0) + 2.0 * np.log(2 * np.pi * 2.0 ** 2)
        )


class TestCompareToManual:
    def test__image(self, gal_data_7x7, sub_mask_7x7):
        galaxy_fit_data = al.GalaxyFitData(
            galaxy_data=gal_data_7x7, mask=sub_mask_7x7, use_image=True
        )

        galaxy = al.Galaxy(
            redshift=0.5,
            light=al.light_profiles.SphericalSersic(centre=(1.0, 2.0), intensity=1.0),
        )
        fit = al.GalaxyFit(galaxy_data=galaxy_fit_data, model_galaxies=[galaxy])

        assert fit.model_galaxies == [galaxy]

        model_data_2d = galaxy.profile_image_from_grid(
            grid=galaxy_fit_data.grid,
            return_in_2d=True,
            return_binned=True,
            bypass_decorator=False,
        )

        residual_map_2d = af.fit_util.residual_map_from_data_mask_and_model_data(
            data=galaxy_fit_data.image(return_in_2d=True),
            mask=galaxy_fit_data.mask,
            model_data=model_data_2d,
        )

        assert residual_map_2d == pytest.approx(
            fit.residual_map(return_in_2d=True), 1e-4
        )

        chi_squared_map_2d = af.fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
            residual_map=residual_map_2d,
            mask=galaxy_fit_data.mask,
            noise_map=galaxy_fit_data.noise_map(return_in_2d=True),
        )

        assert chi_squared_map_2d == pytest.approx(
            fit.chi_squared_map(return_in_2d=True), 1e-4
        )

        chi_squared = af.fit_util.chi_squared_from_chi_squared_map_and_mask(
            chi_squared_map=chi_squared_map_2d, mask=sub_mask_7x7
        )

        noise_normalization = af.fit_util.noise_normalization_from_noise_map_and_mask(
            mask=galaxy_fit_data.mask,
            noise_map=galaxy_fit_data.noise_map(return_in_2d=True),
        )

        likelihood = af.fit_util.likelihood_from_chi_squared_and_noise_normalization(
            chi_squared=chi_squared, noise_normalization=noise_normalization
        )

        assert likelihood == pytest.approx(fit.likelihood, 1e-4)

    def test__convergence(self, gal_data_7x7, sub_mask_7x7):
        galaxy_fit_data = al.GalaxyFitData(
            galaxy_data=gal_data_7x7, mask=sub_mask_7x7, use_convergence=True
        )

        galaxy = al.Galaxy(
            redshift=0.5,
            mass=al.mass_profiles.SphericalIsothermal(
                centre=(1.0, 2.0), einstein_radius=1.0
            ),
        )
        fit = al.GalaxyFit(galaxy_data=galaxy_fit_data, model_galaxies=[galaxy])

        assert fit.model_galaxies == [galaxy]

        model_data_2d = galaxy.convergence_from_grid(
            grid=galaxy_fit_data.grid,
            return_in_2d=True,
            return_binned=True,
            bypass_decorator=False,
        )

        residual_map_2d = af.fit_util.residual_map_from_data_mask_and_model_data(
            data=galaxy_fit_data.image(return_in_2d=True),
            mask=galaxy_fit_data.mask,
            model_data=model_data_2d,
        )
        assert residual_map_2d == pytest.approx(
            fit.residual_map(return_in_2d=True), 1e-4
        )

        chi_squared_map_2d = af.fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
            residual_map=residual_map_2d,
            mask=galaxy_fit_data.mask,
            noise_map=galaxy_fit_data.noise_map(return_in_2d=True),
        )
        assert chi_squared_map_2d == pytest.approx(
            fit.chi_squared_map(return_in_2d=True), 1e-4
        )

        chi_squared = af.fit_util.chi_squared_from_chi_squared_map_and_mask(
            chi_squared_map=chi_squared_map_2d, mask=sub_mask_7x7
        )

        noise_normalization = af.fit_util.noise_normalization_from_noise_map_and_mask(
            mask=galaxy_fit_data.mask,
            noise_map=galaxy_fit_data.noise_map(return_in_2d=True),
        )

        likelihood = af.fit_util.likelihood_from_chi_squared_and_noise_normalization(
            chi_squared=chi_squared, noise_normalization=noise_normalization
        )

        assert likelihood == pytest.approx(fit.likelihood, 1e-4)

    def test__potential(self, gal_data_7x7, sub_mask_7x7):
        galaxy_fit_data = al.GalaxyFitData(
            galaxy_data=gal_data_7x7, mask=sub_mask_7x7, use_potential=True
        )

        galaxy = al.Galaxy(
            redshift=0.5,
            mass=al.mass_profiles.SphericalIsothermal(
                centre=(1.0, 2.0), einstein_radius=1.0
            ),
        )

        fit = al.GalaxyFit(galaxy_data=galaxy_fit_data, model_galaxies=[galaxy])

        assert fit.model_galaxies == [galaxy]

        model_data_2d = galaxy.potential_from_grid(
            grid=galaxy_fit_data.grid,
            return_in_2d=True,
            return_binned=True,
            bypass_decorator=False,
        )

        residual_map_2d = af.fit_util.residual_map_from_data_mask_and_model_data(
            data=galaxy_fit_data.image(return_in_2d=True),
            mask=galaxy_fit_data.mask,
            model_data=model_data_2d,
        )

        assert residual_map_2d == pytest.approx(
            fit.residual_map(return_in_2d=True), 1e-4
        )

        chi_squared_map_2d = af.fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
            residual_map=residual_map_2d,
            mask=galaxy_fit_data.mask,
            noise_map=galaxy_fit_data.noise_map(return_in_2d=True),
        )

        assert chi_squared_map_2d == pytest.approx(
            fit.chi_squared_map(return_in_2d=True), 1e-4
        )

        chi_squared = af.fit_util.chi_squared_from_chi_squared_map_and_mask(
            chi_squared_map=chi_squared_map_2d, mask=sub_mask_7x7
        )

        noise_normalization = af.fit_util.noise_normalization_from_noise_map_and_mask(
            mask=galaxy_fit_data.mask,
            noise_map=galaxy_fit_data.noise_map(return_in_2d=True),
        )

        likelihood = af.fit_util.likelihood_from_chi_squared_and_noise_normalization(
            chi_squared=chi_squared, noise_normalization=noise_normalization
        )

        assert likelihood == pytest.approx(fit.likelihood, 1e-4)

    def test__deflections_y(self, gal_data_7x7, sub_mask_7x7):

        galaxy_fit_data = al.GalaxyFitData(
            galaxy_data=gal_data_7x7, mask=sub_mask_7x7, use_deflections_y=True
        )

        galaxy = al.Galaxy(
            redshift=0.5,
            mass=al.mass_profiles.SphericalIsothermal(
                centre=(1.0, 2.0), einstein_radius=1.0
            ),
        )

        fit = al.GalaxyFit(galaxy_data=galaxy_fit_data, model_galaxies=[galaxy])

        assert fit.model_galaxies == [galaxy]

        model_data_2d = galaxy.deflections_from_grid(
            grid=galaxy_fit_data.grid,
            return_in_2d=True,
            return_binned=True,
            bypass_decorator=False,
        )[:, :, 0]

        residual_map_2d = af.fit_util.residual_map_from_data_mask_and_model_data(
            data=galaxy_fit_data.image(return_in_2d=True),
            mask=galaxy_fit_data.mask,
            model_data=model_data_2d,
        )

        assert residual_map_2d == pytest.approx(
            fit.residual_map(return_in_2d=True), 1e-4
        )

        chi_squared_map_2d = af.fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
            residual_map=residual_map_2d,
            mask=galaxy_fit_data.mask,
            noise_map=galaxy_fit_data.noise_map(return_in_2d=True),
        )

        assert chi_squared_map_2d == pytest.approx(
            fit.chi_squared_map(return_in_2d=True), 1e-4
        )

        chi_squared = af.fit_util.chi_squared_from_chi_squared_map_and_mask(
            chi_squared_map=chi_squared_map_2d, mask=sub_mask_7x7
        )

        noise_normalization = af.fit_util.noise_normalization_from_noise_map_and_mask(
            mask=galaxy_fit_data.mask,
            noise_map=galaxy_fit_data.noise_map(return_in_2d=True),
        )

        likelihood = af.fit_util.likelihood_from_chi_squared_and_noise_normalization(
            chi_squared=chi_squared, noise_normalization=noise_normalization
        )

        assert likelihood == pytest.approx(fit.likelihood, 1e-4)

    def test__deflections_x(self, gal_data_7x7, sub_mask_7x7):

        galaxy_fit_data = al.GalaxyFitData(
            galaxy_data=gal_data_7x7, mask=sub_mask_7x7, use_deflections_x=True
        )

        galaxy = al.Galaxy(
            redshift=0.5,
            mass=al.mass_profiles.SphericalIsothermal(
                centre=(1.0, 2.0), einstein_radius=1.0
            ),
        )
        fit = al.GalaxyFit(galaxy_data=galaxy_fit_data, model_galaxies=[galaxy])

        assert fit.model_galaxies == [galaxy]

        model_data_2d = galaxy.deflections_from_grid(
            grid=galaxy_fit_data.grid,
            return_in_2d=True,
            return_binned=True,
            bypass_decorator=False,
        )[:, :, 1]

        residual_map_2d = af.fit_util.residual_map_from_data_mask_and_model_data(
            data=galaxy_fit_data.image(return_in_2d=True),
            mask=galaxy_fit_data.mask,
            model_data=model_data_2d,
        )

        assert residual_map_2d == pytest.approx(
            fit.residual_map(return_in_2d=True), 1e-4
        )

        chi_squared_map_2d = af.fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
            residual_map=residual_map_2d,
            mask=galaxy_fit_data.mask,
            noise_map=galaxy_fit_data.noise_map(return_in_2d=True),
        )

        assert chi_squared_map_2d == pytest.approx(
            fit.chi_squared_map(return_in_2d=True), 1e-4
        )

        chi_squared = af.fit_util.chi_squared_from_chi_squared_map_and_mask(
            chi_squared_map=chi_squared_map_2d, mask=sub_mask_7x7
        )

        noise_normalization = af.fit_util.noise_normalization_from_noise_map_and_mask(
            mask=galaxy_fit_data.mask,
            noise_map=galaxy_fit_data.noise_map(return_in_2d=True),
        )

        likelihood = af.fit_util.likelihood_from_chi_squared_and_noise_normalization(
            chi_squared=chi_squared, noise_normalization=noise_normalization
        )

        assert likelihood == pytest.approx(fit.likelihood, 1e-4)



import autolens as al
import numpy as np
import pytest

import autofit as af
from autolens.lens.lens_fit.lens_imaging_fit import (
    likelihood_with_regularization_from_chi_squared_regularization_term_and_noise_normalization,
)
from autolens.lens.lens_fit.lens_imaging_fit import evidence_from_inversion_terms
from test.unit.mock.model.mock_profiles import MockLightProfile


class TestInversionEvidence:
    def test__simple_values(self):

        likelihood_with_regularization_terms = likelihood_with_regularization_from_chi_squared_regularization_term_and_noise_normalization(
            chi_squared=3.0, regularization_term=6.0, noise_normalization=2.0
        )

        assert likelihood_with_regularization_terms == -0.5 * (3.0 + 6.0 + 2.0)

        evidences = evidence_from_inversion_terms(
            chi_squared=3.0,
            regularization_term=6.0,
            log_curvature_regularization_term=9.0,
            log_regularization_term=10.0,
            noise_normalization=30.0,
        )

        assert evidences == -0.5 * (3.0 + 6.0 + 9.0 - 10.0 + 30.0)


class TestFitProperties:
    def test__total_inversions(self, lens_imaging_data_7x7):

        g0 = al.Galaxy(redshift=0.5)

        g1 = al.Galaxy(redshift=1.0)

        g2 = al.Galaxy(redshift=2.0)

        tracer = al.Tracer.from_galaxies(galaxies=[g0, g1, g2])

        fit = al.LensImagingFit.from_lens_data_and_tracer(
            lens_data=lens_imaging_data_7x7, tracer=tracer
        )

        assert fit.total_inversions == 0

        g2 = al.Galaxy(
            redshift=2.0,
            pixelization=al.pixelizations.Rectangular(),
            regularization=al.regularization.Constant(),
        )

        tracer = al.Tracer.from_galaxies(galaxies=[g0, g1, g2])

        fit = al.LensImagingFit.from_lens_data_and_tracer(
            lens_data=lens_imaging_data_7x7, tracer=tracer
        )

        assert fit.total_inversions == 1

        g0 = al.Galaxy(
            redshift=0.5,
            pixelization=al.pixelizations.Rectangular(),
            regularization=al.regularization.Constant(),
        )

        g1 = al.Galaxy(
            redshift=1.0,
            pixelization=al.pixelizations.Rectangular(),
            regularization=al.regularization.Constant(),
        )

        g2 = al.Galaxy(
            redshift=2.0,
            pixelization=al.pixelizations.Rectangular(),
            regularization=al.regularization.Constant(),
        )

        tracer = al.Tracer.from_galaxies(galaxies=[g0, g1, g2])

        fit = al.LensImagingFit.from_lens_data_and_tracer(
            lens_data=lens_imaging_data_7x7, tracer=tracer
        )

        assert fit.total_inversions == 3


class TestLikelihood:
    def test__1x2_image__no_psf_blurring__tracing_fits_data_with_chi_sq_5(self):
        # The image plane image generated by the galaxy is [1.0, 1.0]

        # Thus the chi squared is 4.0**2.0 + 3.0**2.0 = 25.0

        psf = al.PSF(
            array=(np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]])),
            pixel_scale=1.0,
        )

        imaging_data = al.ImagingData(
            5.0 * np.ones((3, 4)), pixel_scale=1.0, psf=psf, noise_map=np.ones((3, 4))
        )
        imaging_data.image[1, 2] = 4.0

        mask = al.Mask(
            array=np.array(
                [
                    [True, True, True, True],
                    [True, False, False, True],
                    [True, True, True, True],
                ]
            ),
            pixel_scale=1.0,
            sub_size=1,
        )

        lens_imaging_data_7x7 = al.LensImagingData(imaging_data=imaging_data, mask=mask)

        # Setup as a ray trace instance, using a light profile for the lens

        g0 = al.Galaxy(redshift=0.5, light_profile=MockLightProfile(value=1.0, size=2))
        tracer = al.Tracer.from_galaxies(galaxies=[g0])

        fit = al.LensImagingFit.from_lens_data_and_tracer(
            lens_data=lens_imaging_data_7x7, tracer=tracer
        )

        assert (fit._mask == np.array([False, False])).all()
        assert (
            fit.mask
            == np.array(
                [
                    [True, True, True, True],
                    [True, False, False, True],
                    [True, True, True, True],
                ]
            )
        ).all()

        assert (fit._data == np.array([5.0, 4.0])).all()
        assert (
            fit.image(return_in_2d=True, return_masked=True)
            == np.array(
                [[0.0, 0.0, 0.0, 0.0], [0.0, 5.0, 4.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
            )
        ).all()

        assert (fit._noise_map == np.array([1.0, 1.0])).all()
        assert (
            fit.noise_map(return_in_2d=True, return_masked=True)
            == np.array(
                [[0.0, 0.0, 0.0, 0.0], [0.0, 1.0, 1.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
            )
        ).all()

        assert (fit._model_data == np.array([1.0, 1.0])).all()
        assert (
            fit.model_image(return_in_2d=True)
            == np.array(
                [[0.0, 0.0, 0.0, 0.0], [0.0, 1.0, 1.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
            )
        ).all()

        assert (fit._residual_map == np.array([4.0, 3.0])).all()
        assert (
            fit.residual_map(return_in_2d=True)
            == np.array(
                [[0.0, 0.0, 0.0, 0.0], [0.0, 4.0, 3.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
            )
        ).all()

        assert (fit._normalized_residual_map == np.array([4.0, 3.0])).all()
        assert (
            fit.normalized_residual_map(return_in_2d=True)
            == np.array(
                [[0.0, 0.0, 0.0, 0.0], [0.0, 4.0, 3.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
            )
        ).all()

        assert (fit._chi_squared_map == np.array([16.0, 9.0])).all()
        assert (
            fit.chi_squared_map(return_in_2d=True)
            == np.array(
                [[0.0, 0.0, 0.0, 0.0], [0.0, 16.0, 9.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
            )
        ).all()

        assert fit.chi_squared == 25.0
        assert fit.reduced_chi_squared == 25.0 / 2.0
        assert fit.noise_normalization == (2.0 * np.log(2 * np.pi * 1.0 ** 2.0))
        assert fit.likelihood == -0.5 * (25.0 + 2.0 * np.log(2 * np.pi * 1.0 ** 2.0))

    def test__1x2_image__include_psf_blurring__tracing_fits_data_with_chi_sq_4(self):
        # This PSF changes the blurred image plane image from [1.0, 1.0] to [1.0, 5.0]

        # Thus, the chi squared is 4.0**2.0 + 0.0**2.0 = 16.0

        psf = al.PSF(
            array=(np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 3.0], [0.0, 0.0, 0.0]])),
            pixel_scale=1.0,
        )

        imaging_data = al.ImagingData(
            5.0 * np.ones((3, 4)), pixel_scale=1.0, psf=psf, noise_map=np.ones((3, 4))
        )
        imaging_data.image[1, 2] = 4.0

        mask = al.Mask(
            array=np.array(
                [
                    [True, True, True, True],
                    [True, False, False, True],
                    [True, True, True, True],
                ]
            ),
            pixel_scale=1.0,
            sub_size=1,
        )

        lens_imaging_data_7x7 = al.LensImagingData(imaging_data=imaging_data, mask=mask)

        # Setup as a ray trace instance, using a light profile for the lens

        g0 = al.Galaxy(redshift=0.5, light_profile=MockLightProfile(value=1.0, size=2))
        tracer = al.Tracer.from_galaxies(galaxies=[g0])

        fit = al.LensImagingFit.from_lens_data_and_tracer(
            lens_data=lens_imaging_data_7x7, tracer=tracer
        )

        assert (fit._mask == np.array([False, False])).all()
        assert (
            fit.mask
            == np.array(
                [
                    [True, True, True, True],
                    [True, False, False, True],
                    [True, True, True, True],
                ]
            )
        ).all()

        assert (fit._data == np.array([5.0, 4.0])).all()
        assert (
            fit.image(return_in_2d=True, return_masked=True)
            == np.array(
                [[0.0, 0.0, 0.0, 0.0], [0.0, 5.0, 4.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
            )
        ).all()

        assert (fit._noise_map == np.array([1.0, 1.0])).all()
        assert (
            fit.noise_map(return_in_2d=True, return_masked=True)
            == np.array(
                [[0.0, 0.0, 0.0, 0.0], [0.0, 1.0, 1.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
            )
        ).all()

        assert (fit._model_data == np.array([1.0, 4.0])).all()
        assert (
            fit.model_image(return_in_2d=True)
            == np.array(
                [[0.0, 0.0, 0.0, 0.0], [0.0, 1.0, 4.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
            )
        ).all()

        assert (fit._residual_map == np.array([4.0, 0.0])).all()
        assert (
            fit.residual_map(return_in_2d=True)
            == np.array(
                [[0.0, 0.0, 0.0, 0.0], [0.0, 4.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
            )
        ).all()

        assert (fit._normalized_residual_map == np.array([4.0, 0.0])).all()
        assert (
            fit.normalized_residual_map(return_in_2d=True)
            == np.array(
                [[0.0, 0.0, 0.0, 0.0], [0.0, 4.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
            )
        ).all()

        assert (fit._chi_squared_map == np.array([16.0, 0.0])).all()
        assert (
            fit.chi_squared_map(return_in_2d=True)
            == np.array(
                [[0.0, 0.0, 0.0, 0.0], [0.0, 16.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
            )
        ).all()

        assert fit.chi_squared == 16.0
        assert fit.reduced_chi_squared == 16.0 / 2.0
        assert fit.noise_normalization == (2.0 * np.log(2 * np.pi * 1.0 ** 2.0))
        assert fit.likelihood == -0.5 * (16.0 + 2.0 * np.log(2 * np.pi * 1.0 ** 2.0))

    def test_hyper_galaxy_changes_noise_above_from_1_to_2__reflected_in_likelihood(
        self
    ):
        # This PSF changes the blurred image plane image from [1.0, 1.0] to [1.0, 5.0]

        # Thus, the chi squared is 4.0**2.0 + 0.0**2.0 = 16.0

        # The hyper_galaxies galaxy increases the noise in both pixels by 1.0, to 2.0.

        # This reduces the chi squared to 2.0 instead of 4.0

        psf = al.PSF(
            array=(np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 3.0], [0.0, 0.0, 0.0]])),
            pixel_scale=1.0,
        )

        imaging_data = al.ImagingData(
            5.0 * np.ones((3, 4)), pixel_scale=1.0, psf=psf, noise_map=np.ones((3, 4))
        )
        imaging_data.image[1, 2] = 4.0

        mask = al.Mask(
            array=np.array(
                [
                    [True, True, True, True],
                    [True, False, False, True],
                    [True, True, True, True],
                ]
            ),
            pixel_scale=1.0,
            sub_size=1,
        )

        lens_imaging_data_7x7 = al.LensImagingData(imaging_data=imaging_data, mask=mask)

        # Setup as a ray trace instance, using a light profile for the lens

        g0 = al.Galaxy(
            redshift=0.5,
            light_profile=MockLightProfile(value=1.0, size=2),
            hyper_galaxy=al.HyperGalaxy(
                contribution_factor=1.0, noise_factor=1.0, noise_power=1.0
            ),
            hyper_model_image_1d=np.ones(2),
            hyper_galaxy_image_1d=np.ones(2),
            hyper_minimum_value=0.0,
        )

        tracer = al.Tracer.from_galaxies(galaxies=[g0])

        fit = al.LensImagingFit.from_lens_data_and_tracer(
            lens_data=lens_imaging_data_7x7, tracer=tracer
        )

        assert (fit._mask == np.array([False, False])).all()
        assert (
            fit.mask
            == np.array(
                [
                    [True, True, True, True],
                    [True, False, False, True],
                    [True, True, True, True],
                ]
            )
        ).all()

        assert (fit._data == np.array([5.0, 4.0])).all()
        assert (
            fit.image(return_in_2d=True, return_masked=True)
            == np.array(
                [[0.0, 0.0, 0.0, 0.0], [0.0, 5.0, 4.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
            )
        ).all()

        assert (fit._noise_map == np.array([2.0, 2.0])).all()
        assert (
            fit.noise_map(return_in_2d=True, return_masked=True)
            == np.array(
                [[0.0, 0.0, 0.0, 0.0], [0.0, 2.0, 2.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
            )
        ).all()

        assert (fit._model_data == np.array([1.0, 4.0])).all()
        assert (
            fit.model_image(return_in_2d=True)
            == np.array(
                [[0.0, 0.0, 0.0, 0.0], [0.0, 1.0, 4.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
            )
        ).all()

        assert (fit._residual_map == np.array([4.0, 0.0])).all()
        assert (
            fit.residual_map(return_in_2d=True)
            == np.array(
                [[0.0, 0.0, 0.0, 0.0], [0.0, 4.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
            )
        ).all()

        assert (fit._normalized_residual_map == np.array([2.0, 0.0])).all()
        assert (
            fit.normalized_residual_map(return_in_2d=True)
            == np.array(
                [[0.0, 0.0, 0.0, 0.0], [0.0, 2.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
            )
        ).all()

        assert (fit._chi_squared_map == np.array([4.0, 0.0])).all()
        assert (
            fit.chi_squared_map(return_in_2d=True)
            == np.array(
                [[0.0, 0.0, 0.0, 0.0], [0.0, 4.0, 0.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
            )
        ).all()

        assert fit.chi_squared == 4.0
        assert fit.reduced_chi_squared == 4.0 / 2.0
        assert fit.noise_normalization == (2.0 * np.log(2 * np.pi * 2.0 ** 2.0))
        assert fit.likelihood == -0.5 * (4.0 + 2.0 * np.log(2 * np.pi * 2.0 ** 2.0))

    def test__hyper_image_changes_background_sky__reflected_in_likelihood(self):

        psf = al.PSF(
            array=(np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]])),
            pixel_scale=1.0,
        )

        imaging_data = al.ImagingData(
            5.0 * np.ones((3, 4)), pixel_scale=1.0, psf=psf, noise_map=np.ones((3, 4))
        )
        imaging_data.image[1, 2] = 4.0

        mask = al.Mask(
            array=np.array(
                [
                    [True, True, True, True],
                    [True, False, False, True],
                    [True, True, True, True],
                ]
            ),
            pixel_scale=1.0,
            sub_size=1,
        )

        lens_imaging_data_7x7 = al.LensImagingData(imaging_data=imaging_data, mask=mask)

        # Setup as a ray trace instance, using a light profile for the lens

        g0 = al.Galaxy(redshift=0.5, light_profile=MockLightProfile(value=1.0, size=2))
        tracer = al.Tracer.from_galaxies(galaxies=[g0])

        hyper_image_sky = al.HyperImageSky(sky_scale=1.0)

        fit = al.LensImagingFit.from_lens_data_and_tracer(
            lens_data=lens_imaging_data_7x7,
            tracer=tracer,
            hyper_image_sky=hyper_image_sky,
        )

        assert (fit._mask == np.array([False, False])).all()
        assert (
            fit.mask
            == np.array(
                [
                    [True, True, True, True],
                    [True, False, False, True],
                    [True, True, True, True],
                ]
            )
        ).all()

        assert (fit._data == np.array([6.0, 5.0])).all()
        assert (
            fit.image(return_in_2d=True, return_masked=True)
            == np.array(
                [[0.0, 0.0, 0.0, 0.0], [0.0, 6.0, 5.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
            )
        ).all()

        assert (fit._noise_map == np.array([1.0, 1.0])).all()
        assert (
            fit.noise_map(return_in_2d=True, return_masked=True)
            == np.array(
                [[0.0, 0.0, 0.0, 0.0], [0.0, 1.0, 1.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
            )
        ).all()

        assert (fit._model_data == np.array([1.0, 1.0])).all()
        assert (
            fit.model_image(return_in_2d=True)
            == np.array(
                [[0.0, 0.0, 0.0, 0.0], [0.0, 1.0, 1.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
            )
        ).all()

        assert (fit._residual_map == np.array([5.0, 4.0])).all()
        assert (
            fit.residual_map(return_in_2d=True)
            == np.array(
                [[0.0, 0.0, 0.0, 0.0], [0.0, 5.0, 4.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
            )
        ).all()

        assert (fit._normalized_residual_map == np.array([5.0, 4.0])).all()
        assert (
            fit.normalized_residual_map(return_in_2d=True)
            == np.array(
                [[0.0, 0.0, 0.0, 0.0], [0.0, 5.0, 4.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
            )
        ).all()

        assert (fit._chi_squared_map == np.array([25.0, 16.0])).all()
        assert (
            fit.chi_squared_map(return_in_2d=True)
            == np.array(
                [[0.0, 0.0, 0.0, 0.0], [0.0, 25.0, 16.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
            )
        ).all()

        assert fit.chi_squared == 41.0
        assert fit.reduced_chi_squared == 41.0 / 2.0
        assert fit.noise_normalization == (2.0 * np.log(2 * np.pi * 1.0 ** 2.0))
        assert fit.likelihood == -0.5 * (41.0 + 2.0 * np.log(2 * np.pi * 1.0 ** 2.0))

    def test__hyper_background_changes_background_noise_map__reflected_in_likelihood(
        self
    ):

        psf = al.PSF(
            array=(np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]])),
            pixel_scale=1.0,
        )

        imaging_data = al.ImagingData(
            5.0 * np.ones((3, 4)), pixel_scale=1.0, psf=psf, noise_map=np.ones((3, 4))
        )
        imaging_data.image[1, 2] = 4.0

        mask = al.Mask(
            array=np.array(
                [
                    [True, True, True, True],
                    [True, False, False, True],
                    [True, True, True, True],
                ]
            ),
            pixel_scale=1.0,
            sub_size=1,
        )

        lens_imaging_data_7x7 = al.LensImagingData(imaging_data=imaging_data, mask=mask)

        # Setup as a ray trace instance, using a light profile for the lens

        g0 = al.Galaxy(redshift=0.5, light_profile=MockLightProfile(value=1.0, size=2))
        tracer = al.Tracer.from_galaxies(galaxies=[g0])

        hyper_background_noise = al.HyperBackgroundNoise(noise_scale=1.0)

        fit = al.LensImagingFit.from_lens_data_and_tracer(
            lens_data=lens_imaging_data_7x7,
            tracer=tracer,
            hyper_background_noise=hyper_background_noise,
        )

        assert (fit._mask == np.array([False, False])).all()
        assert (
            fit.mask
            == np.array(
                [
                    [True, True, True, True],
                    [True, False, False, True],
                    [True, True, True, True],
                ]
            )
        ).all()

        assert (fit._data == np.array([5.0, 4.0])).all()
        assert (
            fit.image(return_in_2d=True, return_masked=True)
            == np.array(
                [[0.0, 0.0, 0.0, 0.0], [0.0, 5.0, 4.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
            )
        ).all()

        assert (fit._noise_map == np.array([2.0, 2.0])).all()
        assert (
            fit.noise_map(return_in_2d=True, return_masked=True)
            == np.array(
                [[0.0, 0.0, 0.0, 0.0], [0.0, 2.0, 2.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
            )
        ).all()

        assert (fit._model_data == np.array([1.0, 1.0])).all()
        assert (
            fit.model_image(return_in_2d=True)
            == np.array(
                [[0.0, 0.0, 0.0, 0.0], [0.0, 1.0, 1.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
            )
        ).all()

        assert (fit._residual_map == np.array([4.0, 3.0])).all()
        assert (
            fit.residual_map(return_in_2d=True)
            == np.array(
                [[0.0, 0.0, 0.0, 0.0], [0.0, 4.0, 3.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
            )
        ).all()

        assert (fit._normalized_residual_map == np.array([2.0, 1.5])).all()
        assert (
            fit.normalized_residual_map(return_in_2d=True)
            == np.array(
                [[0.0, 0.0, 0.0, 0.0], [0.0, 2.0, 1.5, 0.0], [0.0, 0.0, 0.0, 0.0]]
            )
        ).all()

        assert (fit._chi_squared_map == np.array([4.0, 2.25])).all()
        assert (
            fit.chi_squared_map(return_in_2d=True)
            == np.array(
                [[0.0, 0.0, 0.0, 0.0], [0.0, 4.0, 2.25, 0.0], [0.0, 0.0, 0.0, 0.0]]
            )
        ).all()

        assert fit.chi_squared == 6.25
        assert fit.reduced_chi_squared == 6.25 / 2.0
        assert fit.noise_normalization == (2.0 * np.log(2 * np.pi * 2.0 ** 2.0))
        assert fit.likelihood == -0.5 * (6.25 + 2.0 * np.log(2 * np.pi * 2.0 ** 2.0))

    def test__hyper_noise_map_max_changes_noise_map__reflected_in_likelihood(self):

        psf = al.PSF(
            array=(np.array([[0.0, 0.0, 0.0], [0.0, 1.0, 0.0], [0.0, 0.0, 0.0]])),
            pixel_scale=1.0,
        )

        imaging_data = al.ImagingData(
            5.0 * np.ones((3, 4)), pixel_scale=1.0, psf=psf, noise_map=np.ones((3, 4))
        )
        imaging_data.image[1, 2] = 4.0

        mask = al.Mask(
            array=np.array(
                [
                    [True, True, True, True],
                    [True, False, False, True],
                    [True, True, True, True],
                ]
            ),
            pixel_scale=1.0,
            sub_size=1,
        )

        lens_imaging_data_7x7 = al.LensImagingData(
            imaging_data=imaging_data, mask=mask, hyper_noise_map_max=1.0
        )

        # Setup as a ray trace instance, using a light profile for the lens

        g0 = al.Galaxy(redshift=0.5, light_profile=MockLightProfile(value=1.0, size=2))
        tracer = al.Tracer.from_galaxies(galaxies=[g0])

        hyper_background_noise = al.HyperBackgroundNoise(noise_scale=1.0)

        fit = al.LensImagingFit.from_lens_data_and_tracer(
            lens_data=lens_imaging_data_7x7,
            tracer=tracer,
            hyper_background_noise=hyper_background_noise,
        )

        assert (fit._mask == np.array([False, False])).all()
        assert (
            fit.mask
            == np.array(
                [
                    [True, True, True, True],
                    [True, False, False, True],
                    [True, True, True, True],
                ]
            )
        ).all()

        assert (fit._data == np.array([5.0, 4.0])).all()
        assert (
            fit.image(return_in_2d=True, return_masked=True)
            == np.array(
                [[0.0, 0.0, 0.0, 0.0], [0.0, 5.0, 4.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
            )
        ).all()

        assert (fit._noise_map == np.array([1.0, 1.0])).all()
        assert (
            fit.noise_map(return_in_2d=True, return_masked=True)
            == np.array(
                [[0.0, 0.0, 0.0, 0.0], [0.0, 1.0, 1.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
            )
        ).all()

        assert (fit._model_data == np.array([1.0, 1.0])).all()
        assert (
            fit.model_image(return_in_2d=True)
            == np.array(
                [[0.0, 0.0, 0.0, 0.0], [0.0, 1.0, 1.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
            )
        ).all()

        assert (fit._residual_map == np.array([4.0, 3.0])).all()
        assert (
            fit.residual_map(return_in_2d=True)
            == np.array(
                [[0.0, 0.0, 0.0, 0.0], [0.0, 4.0, 3.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
            )
        ).all()

        assert (fit._normalized_residual_map == np.array([4.0, 3.0])).all()
        assert (
            fit.normalized_residual_map(return_in_2d=True)
            == np.array(
                [[0.0, 0.0, 0.0, 0.0], [0.0, 4.0, 3.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
            )
        ).all()

        assert (fit._chi_squared_map == np.array([16.0, 9.0])).all()
        assert (
            fit.chi_squared_map(return_in_2d=True)
            == np.array(
                [[0.0, 0.0, 0.0, 0.0], [0.0, 16.0, 9.0, 0.0], [0.0, 0.0, 0.0, 0.0]]
            )
        ).all()

        assert fit.chi_squared == 25.0
        assert fit.reduced_chi_squared == 25.0 / 2.0
        assert fit.noise_normalization == (2.0 * np.log(2 * np.pi * 1.0 ** 2.0))
        assert fit.likelihood == -0.5 * (25.0 + 2.0 * np.log(2 * np.pi * 1.0 ** 2.0))


class TestCompareToManualProfilesOnly:
    def test___all_lens_fit_quantities__no_hyper_methods(self, lens_imaging_data_7x7):

        g0 = al.Galaxy(
            redshift=0.5,
            light_profile=al.light_profiles.EllipticalSersic(intensity=1.0),
            mass_profile=al.mass_profiles.SphericalIsothermal(einstein_radius=1.0),
        )

        g1 = al.Galaxy(
            redshift=1.0,
            light_profile=al.light_profiles.EllipticalSersic(intensity=1.0),
        )

        tracer = al.Tracer.from_galaxies(galaxies=[g0, g1])

        fit = al.LensImagingFit.from_lens_data_and_tracer(
            lens_data=lens_imaging_data_7x7, tracer=tracer
        )

        assert lens_imaging_data_7x7.noise_map(return_in_2d=True) == pytest.approx(
            fit.noise_map(return_in_2d=True, return_masked=True)
        )

        model_image_1d = tracer.blurred_profile_image_from_grid_and_convolver(
            grid=lens_imaging_data_7x7.grid,
            convolver=lens_imaging_data_7x7.convolver,
            blurring_grid=lens_imaging_data_7x7.blurring_grid,
            return_in_2d=False,
        )
        model_image_2d = lens_imaging_data_7x7.mapping.scaled_array_2d_from_array_1d(
            array_1d=model_image_1d
        )

        assert model_image_1d == pytest.approx(fit._model_data, 1e-4)
        assert model_image_2d == pytest.approx(fit.model_image(return_in_2d=True))

        residual_map_1d = af.fit_util.residual_map_from_data_mask_and_model_data(
            data=lens_imaging_data_7x7._image_1d,
            mask=lens_imaging_data_7x7._mask_1d,
            model_data=model_image_1d,
        )
        residual_map_2d = lens_imaging_data_7x7.mapping.scaled_array_2d_from_array_1d(
            array_1d=residual_map_1d
        )

        assert residual_map_1d == pytest.approx(fit._residual_map, 1e-4)
        assert residual_map_2d == pytest.approx(fit.residual_map(return_in_2d=True))

        normalized_residual_map_1d = af.fit_util.normalized_residual_map_from_residual_map_noise_map_and_mask(
            residual_map=residual_map_1d,
            mask=lens_imaging_data_7x7._mask_1d,
            noise_map=lens_imaging_data_7x7._noise_map_1d,
        )

        normalized_residual_map_2d = lens_imaging_data_7x7.mapping.scaled_array_2d_from_array_1d(
            array_1d=normalized_residual_map_1d
        )

        assert normalized_residual_map_1d == pytest.approx(
            fit._normalized_residual_map, 1e-4
        )
        assert normalized_residual_map_2d == pytest.approx(
            fit.normalized_residual_map(return_in_2d=True)
        )

        chi_squared_map_1d = af.fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
            residual_map=residual_map_1d,
            mask=lens_imaging_data_7x7._mask_1d,
            noise_map=lens_imaging_data_7x7._noise_map_1d,
        )

        chi_squared_map_2d = lens_imaging_data_7x7.mapping.scaled_array_2d_from_array_1d(
            array_1d=chi_squared_map_1d
        )

        assert chi_squared_map_1d == pytest.approx(fit._chi_squared_map, 1e-4)
        assert chi_squared_map_2d == pytest.approx(
            fit.chi_squared_map(return_in_2d=True)
        )

        chi_squared = af.fit_util.chi_squared_from_chi_squared_map_and_mask(
            chi_squared_map=chi_squared_map_2d, mask=lens_imaging_data_7x7.mask
        )

        noise_normalization = af.fit_util.noise_normalization_from_noise_map_and_mask(
            noise_map=lens_imaging_data_7x7.noise_map(return_in_2d=True),
            mask=lens_imaging_data_7x7.mask,
        )

        likelihood = af.fit_util.likelihood_from_chi_squared_and_noise_normalization(
            chi_squared=chi_squared, noise_normalization=noise_normalization
        )

        assert likelihood == pytest.approx(fit.likelihood, 1e-4)
        assert likelihood == fit.figure_of_merit

    def test___lens_fit_galaxy_model_image_dict__corresponds_to_blurred_galaxy_images(
        self, lens_imaging_data_7x7
    ):
        g0 = al.Galaxy(
            redshift=0.5,
            light_profile=al.light_profiles.EllipticalSersic(intensity=1.0),
            mass_profile=al.mass_profiles.SphericalIsothermal(einstein_radius=1.0),
        )
        g1 = al.Galaxy(
            redshift=1.0,
            light_profile=al.light_profiles.EllipticalSersic(intensity=1.0),
        )
        g2 = al.Galaxy(redshift=1.0)

        tracer = al.Tracer.from_galaxies(galaxies=[g0, g1, g2])

        fit = al.LensImagingFit.from_lens_data_and_tracer(
            lens_data=lens_imaging_data_7x7, tracer=tracer
        )

        traced_grids_of_planes = tracer.traced_grids_of_planes_from_grid(
            grid=lens_imaging_data_7x7.grid, return_in_2d=False
        )
        traced_blurring_grids_of_planes = tracer.traced_grids_of_planes_from_grid(
            grid=lens_imaging_data_7x7.blurring_grid, return_in_2d=False
        )

        g0_image_1d = g0.profile_image_from_grid(
            grid=traced_grids_of_planes[0], return_in_2d=False, return_binned=True
        )
        g0_blurring_image_1d = g0.profile_image_from_grid(
            grid=traced_blurring_grids_of_planes[0],
            return_in_2d=False,
            return_binned=True,
        )

        g0_blurred_image_1d = lens_imaging_data_7x7.convolver.convolve_image(
            image_array=g0_image_1d, blurring_array=g0_blurring_image_1d
        )

        g1_image_1d = g1.profile_image_from_grid(
            grid=traced_grids_of_planes[1], return_in_2d=False, return_binned=True
        )
        g1_blurring_image_1d = g1.profile_image_from_grid(
            grid=traced_blurring_grids_of_planes[1],
            return_in_2d=False,
            return_binned=True,
        )

        g1_blurred_image_1d = lens_imaging_data_7x7.convolver.convolve_image(
            image_array=g1_image_1d, blurring_array=g1_blurring_image_1d
        )

        assert fit.galaxy_model_image_1d_dict[g0] == pytest.approx(
            g0_blurred_image_1d, 1.0e-4
        )
        assert fit.galaxy_model_image_1d_dict[g1] == pytest.approx(
            g1_blurred_image_1d, 1.0e-4
        )
        assert (fit.galaxy_model_image_1d_dict[g2] == np.zeros(9)).all()

        assert fit._model_data == pytest.approx(
            fit.galaxy_model_image_1d_dict[g0] + fit.galaxy_model_image_1d_dict[g1],
            1.0e-4,
        )

        g0_blurred_image_2d = lens_imaging_data_7x7.mapping.scaled_array_2d_from_array_1d(
            array_1d=g0_blurred_image_1d
        )
        g1_blurred_image_2d = lens_imaging_data_7x7.mapping.scaled_array_2d_from_array_1d(
            array_1d=g1_blurred_image_1d
        )

        assert fit.galaxy_model_image_2d_dict[g0] == pytest.approx(
            g0_blurred_image_2d, 1.0e-4
        )
        assert fit.galaxy_model_image_2d_dict[g1] == pytest.approx(
            g1_blurred_image_2d, 1.0e-4
        )

        assert fit.model_image(return_in_2d=True) == pytest.approx(
            fit.galaxy_model_image_2d_dict[g0] + fit.galaxy_model_image_2d_dict[g1],
            1.0e-4,
        )

    def test___all_lens_fit_quantities__including_hyper_methods(
        self, lens_imaging_data_7x7
    ):

        hyper_noise_map_max = 0.2
        lens_imaging_data_7x7.hyper_noise_map_max = hyper_noise_map_max

        hyper_image_sky = al.HyperImageSky(sky_scale=1.0)

        hyper_background_noise = al.HyperBackgroundNoise(noise_scale=1.0)

        image_1d = hyper_image_sky.image_scaled_sky_from_image(
            image=lens_imaging_data_7x7._image_1d
        )

        hyper_noise_map_background_1d = hyper_background_noise.noise_map_scaled_noise_from_noise_map(
            noise_map=lens_imaging_data_7x7._noise_map_1d
        )

        g0 = al.Galaxy(
            redshift=0.5,
            light_profile=al.light_profiles.EllipticalSersic(intensity=1.0),
            mass_profile=al.mass_profiles.SphericalIsothermal(einstein_radius=1.0),
            hyper_galaxy=al.HyperGalaxy(
                contribution_factor=1.0, noise_factor=1.0, noise_power=1.0
            ),
            hyper_model_image_1d=np.ones(9),
            hyper_galaxy_image_1d=np.ones(9),
            hyper_minimum_value=0.0,
        )
        g1 = al.Galaxy(
            redshift=1.0,
            light_profile=al.light_profiles.EllipticalSersic(intensity=1.0),
        )

        tracer = al.Tracer.from_galaxies(galaxies=[g0, g1])

        fit = al.LensImagingFit.from_lens_data_and_tracer(
            lens_data=lens_imaging_data_7x7,
            tracer=tracer,
            hyper_image_sky=hyper_image_sky,
            hyper_background_noise=hyper_background_noise,
        )

        hyper_noise_1d = tracer.hyper_noise_map_1d_from_noise_map_1d(
            noise_map_1d=lens_imaging_data_7x7._noise_map_1d
        )

        hyper_noise_map_1d = hyper_noise_map_background_1d + hyper_noise_1d

        hyper_noise_map_1d[
            hyper_noise_map_1d > hyper_noise_map_max
        ] = hyper_noise_map_max

        hyper_noise_map_2d = lens_imaging_data_7x7.mapping.scaled_array_2d_from_array_1d(
            array_1d=hyper_noise_map_1d
        )

        assert hyper_noise_map_1d == pytest.approx(fit._noise_map, 1e-4)
        assert hyper_noise_map_2d == pytest.approx(
            fit.noise_map(return_in_2d=True, return_masked=True)
        )

        model_image_1d = tracer.blurred_profile_image_from_grid_and_convolver(
            grid=lens_imaging_data_7x7.grid,
            convolver=lens_imaging_data_7x7.convolver,
            blurring_grid=lens_imaging_data_7x7.blurring_grid,
            return_in_2d=False,
        )

        model_image_2d = lens_imaging_data_7x7.mapping.scaled_array_2d_from_array_1d(
            array_1d=model_image_1d
        )

        assert model_image_1d == pytest.approx(fit._model_data, 1e-4)
        assert model_image_2d == pytest.approx(fit.model_image(return_in_2d=True))

        residual_map_1d = af.fit_util.residual_map_from_data_mask_and_model_data(
            data=image_1d,
            mask=lens_imaging_data_7x7._mask_1d,
            model_data=model_image_1d,
        )
        residual_map_2d = lens_imaging_data_7x7.mapping.scaled_array_2d_from_array_1d(
            array_1d=residual_map_1d
        )

        assert residual_map_1d == pytest.approx(fit._residual_map, 1e-4)
        assert residual_map_2d == pytest.approx(fit.residual_map(return_in_2d=True))

        normalized_residual_map_1d = af.fit_util.normalized_residual_map_from_residual_map_noise_map_and_mask(
            residual_map=residual_map_1d,
            mask=lens_imaging_data_7x7._mask_1d,
            noise_map=hyper_noise_map_1d,
        )

        normalized_residual_map_2d = lens_imaging_data_7x7.mapping.scaled_array_2d_from_array_1d(
            array_1d=normalized_residual_map_1d
        )

        assert normalized_residual_map_1d == pytest.approx(
            fit._normalized_residual_map, 1e-4
        )
        assert normalized_residual_map_2d == pytest.approx(
            fit.normalized_residual_map(return_in_2d=True)
        )

        chi_squared_map_1d = af.fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
            residual_map=residual_map_1d,
            mask=lens_imaging_data_7x7._mask_1d,
            noise_map=hyper_noise_map_1d,
        )

        chi_squared_map_2d = lens_imaging_data_7x7.mapping.scaled_array_2d_from_array_1d(
            array_1d=chi_squared_map_1d
        )

        assert chi_squared_map_1d == pytest.approx(fit._chi_squared_map, 1e-4)
        assert chi_squared_map_2d == pytest.approx(
            fit.chi_squared_map(return_in_2d=True)
        )

        chi_squared = af.fit_util.chi_squared_from_chi_squared_map_and_mask(
            chi_squared_map=chi_squared_map_2d, mask=lens_imaging_data_7x7.mask
        )

        noise_normalization = af.fit_util.noise_normalization_from_noise_map_and_mask(
            noise_map=hyper_noise_map_2d, mask=lens_imaging_data_7x7.mask
        )

        likelihood = af.fit_util.likelihood_from_chi_squared_and_noise_normalization(
            chi_squared=chi_squared, noise_normalization=noise_normalization
        )

        assert likelihood == pytest.approx(fit.likelihood, 1e-4)
        assert likelihood == fit.figure_of_merit

    def test___blurred_and_model_images_of_planes_and_unmasked_blurred_profile_image_properties(
        self, lens_imaging_data_7x7
    ):

        g0 = al.Galaxy(
            redshift=0.5,
            light_profile=al.light_profiles.EllipticalSersic(intensity=1.0),
            mass_profile=al.mass_profiles.SphericalIsothermal(einstein_radius=1.0),
        )

        g1 = al.Galaxy(
            redshift=1.0,
            light_profile=al.light_profiles.EllipticalSersic(intensity=1.0),
        )

        tracer = al.Tracer.from_galaxies(galaxies=[g0, g1])

        fit = al.LensImagingFit.from_lens_data_and_tracer(
            lens_data=lens_imaging_data_7x7, tracer=tracer
        )

        blurred_profile_image_2d_of_planes = tracer.blurred_profile_images_of_planes_from_grid_and_convolver(
            grid=lens_imaging_data_7x7.grid,
            convolver=lens_imaging_data_7x7.convolver,
            blurring_grid=lens_imaging_data_7x7.blurring_grid,
            return_in_2d=True,
        )

        model_images_of_planes = fit.model_images_of_planes(return_in_2d=True)

        assert blurred_profile_image_2d_of_planes[0] == pytest.approx(
            model_images_of_planes[0], 1.0e-4
        )
        assert blurred_profile_image_2d_of_planes[1] == pytest.approx(
            model_images_of_planes[1], 1.0e-4
        )

        unmasked_blurred_profile_image = tracer.unmasked_blurred_profile_image_from_grid_and_psf(
            grid=lens_imaging_data_7x7.grid, psf=lens_imaging_data_7x7.psf
        )

        assert (
            unmasked_blurred_profile_image == fit.unmasked_blurred_profile_image
        ).all()

        unmasked_blurred_profile_image_of_planes = tracer.unmasked_blurred_profile_image_of_planes_from_grid_and_psf(
            grid=lens_imaging_data_7x7.grid, psf=lens_imaging_data_7x7.psf
        )

        assert (
            unmasked_blurred_profile_image_of_planes[0]
            == fit.unmasked_blurred_profile_image_of_planes[0]
        ).all()
        assert (
            unmasked_blurred_profile_image_of_planes[1]
            == fit.unmasked_blurred_profile_image_of_planes[1]
        ).all()

        unmasked_blurred_profile_image_of_galaxies = tracer.unmasked_blurred_profile_image_of_planes_and_galaxies_from_grid_and_psf(
            grid=lens_imaging_data_7x7.grid, psf=lens_imaging_data_7x7.psf
        )

        assert (
            unmasked_blurred_profile_image_of_galaxies[0][0]
            == fit.unmasked_blurred_profile_image_of_planes_and_galaxies[0][0]
        ).all()
        assert (
            unmasked_blurred_profile_image_of_galaxies[1][0]
            == fit.unmasked_blurred_profile_image_of_planes_and_galaxies[1][0]
        ).all()


class TestCompareToManualInversionOnly:
    def test___all_lens_fit_quantities__no_hyper_methods(self, lens_imaging_data_7x7):

        pix = al.pixelizations.Rectangular(shape=(3, 3))
        reg = al.regularization.Constant(coefficient=1.0)

        g0 = al.Galaxy(redshift=0.5, pixelization=pix, regularization=reg)

        tracer = al.Tracer.from_galaxies(galaxies=[al.Galaxy(redshift=0.5), g0])

        fit = al.LensImagingFit.from_lens_data_and_tracer(
            lens_data=lens_imaging_data_7x7, tracer=tracer
        )

        mapper = pix.mapper_from_grid_and_pixelization_grid(
            grid=lens_imaging_data_7x7.grid, pixelization_grid=None
        )
        inversion = al.Inversion.from_data_1d_mapper_and_regularization(
            mapper=mapper,
            regularization=reg,
            image_1d=lens_imaging_data_7x7._image_1d,
            noise_map_1d=lens_imaging_data_7x7._noise_map_1d,
            convolver=lens_imaging_data_7x7.convolver,
        )

        assert inversion.reconstructed_data_1d == pytest.approx(fit._model_data, 1e-4)
        assert inversion.reconstructed_data_2d == fit.model_image(return_in_2d=True)

        residual_map_1d = af.fit_util.residual_map_from_data_mask_and_model_data(
            data=lens_imaging_data_7x7._image_1d,
            mask=lens_imaging_data_7x7._mask_1d,
            model_data=inversion.reconstructed_data_1d,
        )
        residual_map_2d = lens_imaging_data_7x7.mapping.scaled_array_2d_from_array_1d(
            array_1d=residual_map_1d
        )

        assert residual_map_1d == pytest.approx(fit._residual_map, 1e-4)
        assert residual_map_2d == pytest.approx(fit.residual_map(return_in_2d=True))

        normalized_residual_map_1d = af.fit_util.normalized_residual_map_from_residual_map_noise_map_and_mask(
            residual_map=residual_map_1d,
            mask=lens_imaging_data_7x7._mask_1d,
            noise_map=lens_imaging_data_7x7._noise_map_1d,
        )

        normalized_residual_map_2d = lens_imaging_data_7x7.mapping.scaled_array_2d_from_array_1d(
            array_1d=normalized_residual_map_1d
        )

        assert normalized_residual_map_1d == pytest.approx(
            fit._normalized_residual_map, 1e-4
        )
        assert normalized_residual_map_2d == pytest.approx(
            fit.normalized_residual_map(return_in_2d=True)
        )

        chi_squared_map_1d = af.fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
            residual_map=residual_map_1d,
            mask=lens_imaging_data_7x7._mask_1d,
            noise_map=lens_imaging_data_7x7._noise_map_1d,
        )
        chi_squared_map_2d = lens_imaging_data_7x7.mapping.scaled_array_2d_from_array_1d(
            array_1d=chi_squared_map_1d
        )

        assert chi_squared_map_1d == pytest.approx(fit._chi_squared_map, 1e-4)
        assert chi_squared_map_2d == pytest.approx(
            fit.chi_squared_map(return_in_2d=True)
        )

        chi_squared = af.fit_util.chi_squared_from_chi_squared_map_and_mask(
            chi_squared_map=chi_squared_map_2d, mask=lens_imaging_data_7x7.mask
        )

        noise_normalization = af.fit_util.noise_normalization_from_noise_map_and_mask(
            mask=lens_imaging_data_7x7.mask,
            noise_map=lens_imaging_data_7x7.noise_map(return_in_2d=True),
        )

        likelihood = af.fit_util.likelihood_from_chi_squared_and_noise_normalization(
            chi_squared=chi_squared, noise_normalization=noise_normalization
        )

        assert likelihood == pytest.approx(fit.likelihood, 1e-4)

        likelihood_with_regularization = likelihood_with_regularization_from_chi_squared_regularization_term_and_noise_normalization(
            chi_squared=chi_squared,
            regularization_term=inversion.regularization_term,
            noise_normalization=noise_normalization,
        )

        assert likelihood_with_regularization == pytest.approx(
            fit.likelihood_with_regularization, 1e-4
        )

        evidence = evidence_from_inversion_terms(
            chi_squared=chi_squared,
            regularization_term=inversion.regularization_term,
            log_curvature_regularization_term=inversion.log_det_curvature_reg_matrix_term,
            log_regularization_term=inversion.log_det_regularization_matrix_term,
            noise_normalization=noise_normalization,
        )

        assert evidence == fit.evidence
        assert evidence == fit.figure_of_merit

    def test___lens_fit_galaxy_model_image_dict__has_inversion_reconstructed_data(
        self, lens_imaging_data_7x7
    ):
        pix = al.pixelizations.Rectangular(shape=(3, 3))
        reg = al.regularization.Constant(coefficient=1.0)

        g0 = al.Galaxy(redshift=0.5)
        g1 = al.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)

        tracer = al.Tracer.from_galaxies(galaxies=[g0, g1])

        fit = al.LensImagingFit.from_lens_data_and_tracer(
            lens_data=lens_imaging_data_7x7, tracer=tracer
        )

        mapper = pix.mapper_from_grid_and_pixelization_grid(
            grid=lens_imaging_data_7x7.grid, pixelization_grid=None
        )

        inversion = al.Inversion.from_data_1d_mapper_and_regularization(
            mapper=mapper,
            regularization=reg,
            image_1d=lens_imaging_data_7x7._image_1d,
            noise_map_1d=lens_imaging_data_7x7._noise_map_1d,
            convolver=lens_imaging_data_7x7.convolver,
        )

        assert (fit.galaxy_model_image_1d_dict[g0] == np.zeros(9)).all()

        assert fit.galaxy_model_image_1d_dict[g1] == pytest.approx(
            inversion.reconstructed_data_1d, 1.0e-4
        )

        assert fit._model_data == pytest.approx(
            fit.galaxy_model_image_1d_dict[g1], 1.0e-4
        )

        assert fit.galaxy_model_image_2d_dict[g1] == pytest.approx(
            inversion.reconstructed_data_2d, 1.0e-4
        )

        assert fit.model_image(return_in_2d=True) == pytest.approx(
            fit.galaxy_model_image_2d_dict[g1], 1.0e-4
        )

    def test___all_lens_fit_quantities__include_hyper_methods(
        self, lens_imaging_data_7x7
    ):

        hyper_noise_map_max = 0.2
        lens_imaging_data_7x7.hyper_noise_map_max = hyper_noise_map_max

        hyper_image_sky = al.HyperImageSky(sky_scale=1.0)

        hyper_background_noise = al.HyperBackgroundNoise(noise_scale=1.0)

        image_1d = hyper_image_sky.image_scaled_sky_from_image(
            image=lens_imaging_data_7x7._image_1d
        )

        hyper_noise_map_background_1d = hyper_background_noise.noise_map_scaled_noise_from_noise_map(
            noise_map=lens_imaging_data_7x7._noise_map_1d
        )

        pix = al.pixelizations.Rectangular(shape=(3, 3))
        reg = al.regularization.Constant(coefficient=1.0)

        g0 = al.Galaxy(
            redshift=0.5,
            pixelization=pix,
            regularization=reg,
            hyper_galaxy=al.HyperGalaxy(
                contribution_factor=1.0, noise_factor=1.0, noise_power=1.0
            ),
            hyper_model_image_1d=np.ones(9),
            hyper_galaxy_image_1d=np.ones(9),
            hyper_minimum_value=0.0,
        )

        tracer = al.Tracer.from_galaxies(galaxies=[al.Galaxy(redshift=0.5), g0])

        fit = al.LensImagingFit.from_lens_data_and_tracer(
            lens_data=lens_imaging_data_7x7,
            tracer=tracer,
            hyper_image_sky=hyper_image_sky,
            hyper_background_noise=hyper_background_noise,
        )

        hyper_noise_1d = tracer.hyper_noise_map_1d_from_noise_map_1d(
            noise_map_1d=lens_imaging_data_7x7._noise_map_1d
        )
        hyper_noise_map_1d = hyper_noise_map_background_1d + hyper_noise_1d
        hyper_noise_map_1d[
            hyper_noise_map_1d > hyper_noise_map_max
        ] = hyper_noise_map_max

        hyper_noise_map_2d = lens_imaging_data_7x7.mapping.scaled_array_2d_from_array_1d(
            array_1d=hyper_noise_map_1d
        )
        assert hyper_noise_map_1d == pytest.approx(fit._noise_map, 1e-4)
        assert hyper_noise_map_2d == pytest.approx(
            fit.noise_map(return_in_2d=True, return_masked=True)
        )

        mapper = pix.mapper_from_grid_and_pixelization_grid(
            grid=lens_imaging_data_7x7.grid, inversion_uses_border=False
        )
        inversion = al.Inversion.from_data_1d_mapper_and_regularization(
            mapper=mapper,
            regularization=reg,
            image_1d=image_1d,
            noise_map_1d=hyper_noise_map_1d,
            convolver=lens_imaging_data_7x7.convolver,
        )

        assert inversion.reconstructed_data_1d == pytest.approx(fit._model_data, 1e-4)
        assert inversion.reconstructed_data_2d == fit.model_image(return_in_2d=True)

        residual_map_1d = af.fit_util.residual_map_from_data_mask_and_model_data(
            data=image_1d,
            mask=lens_imaging_data_7x7._mask_1d,
            model_data=inversion.reconstructed_data_1d,
        )
        residual_map_2d = lens_imaging_data_7x7.mapping.scaled_array_2d_from_array_1d(
            array_1d=residual_map_1d
        )

        assert residual_map_1d == pytest.approx(fit._residual_map, 1e-4)
        assert residual_map_2d == pytest.approx(fit.residual_map(return_in_2d=True))

        normalized_residual_map_1d = af.fit_util.normalized_residual_map_from_residual_map_noise_map_and_mask(
            residual_map=residual_map_1d,
            mask=lens_imaging_data_7x7._mask_1d,
            noise_map=hyper_noise_map_1d,
        )

        normalized_residual_map_2d = lens_imaging_data_7x7.mapping.scaled_array_2d_from_array_1d(
            array_1d=normalized_residual_map_1d
        )

        assert normalized_residual_map_1d == pytest.approx(
            fit._normalized_residual_map, 1e-4
        )
        assert normalized_residual_map_2d == pytest.approx(
            fit.normalized_residual_map(return_in_2d=True)
        )

        chi_squared_map_1d = af.fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
            residual_map=residual_map_1d,
            mask=lens_imaging_data_7x7._mask_1d,
            noise_map=hyper_noise_map_1d,
        )
        chi_squared_map_2d = lens_imaging_data_7x7.mapping.scaled_array_2d_from_array_1d(
            array_1d=chi_squared_map_1d
        )

        assert chi_squared_map_1d == pytest.approx(fit._chi_squared_map, 1e-4)
        assert chi_squared_map_2d == pytest.approx(
            fit.chi_squared_map(return_in_2d=True)
        )

        chi_squared = af.fit_util.chi_squared_from_chi_squared_map_and_mask(
            chi_squared_map=chi_squared_map_2d, mask=lens_imaging_data_7x7.mask
        )

        noise_normalization = af.fit_util.noise_normalization_from_noise_map_and_mask(
            mask=lens_imaging_data_7x7.mask, noise_map=hyper_noise_map_2d
        )

        likelihood = af.fit_util.likelihood_from_chi_squared_and_noise_normalization(
            chi_squared=chi_squared, noise_normalization=noise_normalization
        )

        assert likelihood == pytest.approx(fit.likelihood, 1e-4)

        likelihood_with_regularization = likelihood_with_regularization_from_chi_squared_regularization_term_and_noise_normalization(
            chi_squared=chi_squared,
            regularization_term=inversion.regularization_term,
            noise_normalization=noise_normalization,
        )

        assert likelihood_with_regularization == pytest.approx(
            fit.likelihood_with_regularization, 1e-4
        )

        evidence = evidence_from_inversion_terms(
            chi_squared=chi_squared,
            regularization_term=inversion.regularization_term,
            log_curvature_regularization_term=inversion.log_det_curvature_reg_matrix_term,
            log_regularization_term=inversion.log_det_regularization_matrix_term,
            noise_normalization=noise_normalization,
        )

        assert evidence == fit.evidence
        assert evidence == fit.figure_of_merit

    def test___blurred_and_model_images_of_planes_and_unmasked_blurred_profile_image_properties(
        self, lens_imaging_data_7x7
    ):

        pix = al.pixelizations.Rectangular(shape=(3, 3))
        reg = al.regularization.Constant(coefficient=1.0)

        g0 = al.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)

        tracer = al.Tracer.from_galaxies(galaxies=[al.Galaxy(redshift=0.5), g0])

        fit = al.LensImagingFit.from_lens_data_and_tracer(
            lens_data=lens_imaging_data_7x7, tracer=tracer
        )

        mapper = pix.mapper_from_grid_and_pixelization_grid(
            grid=lens_imaging_data_7x7.grid, inversion_uses_border=False
        )

        inversion = al.Inversion.from_data_1d_mapper_and_regularization(
            mapper=mapper,
            regularization=reg,
            image_1d=lens_imaging_data_7x7._image_1d,
            noise_map_1d=lens_imaging_data_7x7._noise_map_1d,
            convolver=lens_imaging_data_7x7.convolver,
        )

        model_images_of_planes = fit.model_images_of_planes(return_in_2d=True)

        assert (model_images_of_planes[0] == np.zeros((7, 7))).all()
        assert inversion.reconstructed_data_2d == pytest.approx(
            model_images_of_planes[1], 1.0e-4
        )


class TestCompareToManualProfilesAndInversion:
    def test___all_lens_fit_quantities__no_hyper_methods(self, lens_imaging_data_7x7):
        galaxy_light = al.Galaxy(
            redshift=0.5,
            light_profile=al.light_profiles.EllipticalSersic(intensity=1.0),
        )

        pix = al.pixelizations.Rectangular(shape=(3, 3))
        reg = al.regularization.Constant(coefficient=1.0)
        galaxy_pix = al.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)

        tracer = al.Tracer.from_galaxies(galaxies=[galaxy_light, galaxy_pix])

        fit = al.LensImagingFit.from_lens_data_and_tracer(
            lens_data=lens_imaging_data_7x7, tracer=tracer
        )

        blurred_profile_image_1d = tracer.blurred_profile_image_from_grid_and_convolver(
            grid=lens_imaging_data_7x7.grid,
            convolver=lens_imaging_data_7x7.convolver,
            blurring_grid=lens_imaging_data_7x7.blurring_grid,
            return_in_2d=False,
        )

        blurred_profile_image_2d = lens_imaging_data_7x7.mapping.scaled_array_2d_from_array_1d(
            array_1d=blurred_profile_image_1d
        )

        assert blurred_profile_image_1d == pytest.approx(
            fit.blurred_profile_image(return_in_2d=False), 1e-4
        )
        assert blurred_profile_image_2d == pytest.approx(
            fit.blurred_profile_image(return_in_2d=True)
        )

        profile_subtracted_image_1d = (
            lens_imaging_data_7x7._image_1d - blurred_profile_image_1d
        )
        profile_subtracted_image_2d = lens_imaging_data_7x7.mapping.scaled_array_2d_from_array_1d(
            array_1d=profile_subtracted_image_1d
        )

        assert profile_subtracted_image_1d == pytest.approx(
            fit.profile_subtracted_image(return_in_2d=False)
        )
        assert profile_subtracted_image_2d == pytest.approx(
            fit.profile_subtracted_image(return_in_2d=True)
        )

        mapper = pix.mapper_from_grid_and_pixelization_grid(
            grid=lens_imaging_data_7x7.grid, inversion_uses_border=False
        )

        inversion = al.Inversion.from_data_1d_mapper_and_regularization(
            image_1d=profile_subtracted_image_1d,
            noise_map_1d=lens_imaging_data_7x7._noise_map_1d,
            convolver=lens_imaging_data_7x7.convolver,
            mapper=mapper,
            regularization=reg,
        )

        model_image_1d = blurred_profile_image_1d + inversion.reconstructed_data_1d
        model_image_2d = lens_imaging_data_7x7.mapping.scaled_array_2d_from_array_1d(
            array_1d=model_image_1d
        )

        assert model_image_1d == pytest.approx(fit._model_data, 1e-4)
        assert model_image_2d == pytest.approx(fit.model_image(return_in_2d=True))

        residual_map_1d = af.fit_util.residual_map_from_data_mask_and_model_data(
            data=lens_imaging_data_7x7._image_1d,
            mask=lens_imaging_data_7x7._mask_1d,
            model_data=model_image_1d,
        )
        residual_map_2d = lens_imaging_data_7x7.mapping.scaled_array_2d_from_array_1d(
            array_1d=residual_map_1d
        )

        assert residual_map_1d == pytest.approx(fit._residual_map, 1e-4)
        assert residual_map_2d == pytest.approx(fit.residual_map(return_in_2d=True))

        normalized_residual_map_1d = af.fit_util.normalized_residual_map_from_residual_map_noise_map_and_mask(
            residual_map=residual_map_1d,
            mask=lens_imaging_data_7x7._mask_1d,
            noise_map=lens_imaging_data_7x7._noise_map_1d,
        )

        normalized_residual_map_2d = lens_imaging_data_7x7.mapping.scaled_array_2d_from_array_1d(
            array_1d=normalized_residual_map_1d
        )

        assert normalized_residual_map_1d == pytest.approx(
            fit._normalized_residual_map, 1e-4
        )
        assert normalized_residual_map_2d == pytest.approx(
            fit.normalized_residual_map(return_in_2d=True)
        )

        chi_squared_map_1d = af.fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
            residual_map=residual_map_1d,
            mask=lens_imaging_data_7x7._mask_1d,
            noise_map=lens_imaging_data_7x7._noise_map_1d,
        )
        chi_squared_map_2d = lens_imaging_data_7x7.mapping.scaled_array_2d_from_array_1d(
            array_1d=chi_squared_map_1d
        )

        assert chi_squared_map_1d == pytest.approx(fit._chi_squared_map, 1e-4)
        assert chi_squared_map_2d == pytest.approx(
            fit.chi_squared_map(return_in_2d=True)
        )

        chi_squared = af.fit_util.chi_squared_from_chi_squared_map_and_mask(
            chi_squared_map=chi_squared_map_2d, mask=lens_imaging_data_7x7.mask
        )

        noise_normalization = af.fit_util.noise_normalization_from_noise_map_and_mask(
            mask=lens_imaging_data_7x7.mask,
            noise_map=lens_imaging_data_7x7.noise_map(return_in_2d=True),
        )

        likelihood = af.fit_util.likelihood_from_chi_squared_and_noise_normalization(
            chi_squared=chi_squared, noise_normalization=noise_normalization
        )

        assert likelihood == pytest.approx(fit.likelihood, 1e-4)

        likelihood_with_regularization = likelihood_with_regularization_from_chi_squared_regularization_term_and_noise_normalization(
            chi_squared=chi_squared,
            regularization_term=inversion.regularization_term,
            noise_normalization=noise_normalization,
        )

        assert likelihood_with_regularization == pytest.approx(
            fit.likelihood_with_regularization, 1e-4
        )

        evidence = evidence_from_inversion_terms(
            chi_squared=chi_squared,
            regularization_term=inversion.regularization_term,
            log_curvature_regularization_term=inversion.log_det_curvature_reg_matrix_term,
            log_regularization_term=inversion.log_det_regularization_matrix_term,
            noise_normalization=noise_normalization,
        )

        assert evidence == fit.evidence
        assert evidence == fit.figure_of_merit

    def test___lens_fit_galaxy_model_image_dict__has_blurred_profile_images_and_inversion_reconstructed_data(
        self, lens_imaging_data_7x7
    ):

        g0 = al.Galaxy(
            redshift=0.5,
            light_profile=al.light_profiles.EllipticalSersic(intensity=1.0),
        )
        g1 = al.Galaxy(
            redshift=0.5,
            light_profile=al.light_profiles.EllipticalSersic(intensity=2.0),
        )
        g2 = al.Galaxy(redshift=0.5)

        pix = al.pixelizations.Rectangular(shape=(3, 3))
        reg = al.regularization.Constant(coefficient=1.0)
        galaxy_pix = al.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)

        tracer = al.Tracer.from_galaxies(galaxies=[g0, g1, g2, galaxy_pix])

        fit = al.LensImagingFit.from_lens_data_and_tracer(
            lens_data=lens_imaging_data_7x7, tracer=tracer
        )

        traced_grids = tracer.traced_grids_of_planes_from_grid(
            grid=lens_imaging_data_7x7.grid
        )
        traced_blurring_grids = tracer.traced_grids_of_planes_from_grid(
            grid=lens_imaging_data_7x7.blurring_grid
        )

        g0_image_1d = g0.profile_image_from_grid(
            grid=traced_grids[0], return_in_2d=False, return_binned=True
        )
        g0_blurring_image_1d = g0.profile_image_from_grid(
            grid=traced_blurring_grids[0], return_in_2d=False, return_binned=False
        )

        g0_blurred_image_1d = lens_imaging_data_7x7.convolver.convolve_image(
            image_array=g0_image_1d, blurring_array=g0_blurring_image_1d
        )

        g1_image_1d = g1.profile_image_from_grid(
            grid=traced_grids[1], return_in_2d=False, return_binned=True
        )
        g1_blurring_image_1d = g1.profile_image_from_grid(
            grid=traced_blurring_grids[1], return_in_2d=False, return_binned=False
        )

        g1_blurred_image_1d = lens_imaging_data_7x7.convolver.convolve_image(
            image_array=g1_image_1d, blurring_array=g1_blurring_image_1d
        )

        blurred_profile_image_1d = g0_blurred_image_1d + g1_blurred_image_1d

        profile_subtracted_image_1d = (
            lens_imaging_data_7x7._image_1d - blurred_profile_image_1d
        )
        mapper = pix.mapper_from_grid_and_pixelization_grid(
            grid=lens_imaging_data_7x7.grid, inversion_uses_border=False
        )

        inversion = al.Inversion.from_data_1d_mapper_and_regularization(
            image_1d=profile_subtracted_image_1d,
            noise_map_1d=lens_imaging_data_7x7._noise_map_1d,
            convolver=lens_imaging_data_7x7.convolver,
            mapper=mapper,
            regularization=reg,
        )

        assert fit.galaxy_model_image_1d_dict[g0] == pytest.approx(
            g0_blurred_image_1d, 1.0e-4
        )
        assert fit.galaxy_model_image_1d_dict[g1] == pytest.approx(
            g1_blurred_image_1d, 1.0e-4
        )
        assert (fit.galaxy_model_image_1d_dict[g2] == np.zeros(9)).all()
        assert fit.galaxy_model_image_1d_dict[galaxy_pix] == pytest.approx(
            inversion.reconstructed_data_1d, 1.0e-4
        )

        assert fit._model_data == pytest.approx(
            fit.galaxy_model_image_1d_dict[g0]
            + fit.galaxy_model_image_1d_dict[g1]
            + fit.galaxy_model_image_1d_dict[galaxy_pix],
            1.0e-4,
        )

        g0_blurred_image_2d = lens_imaging_data_7x7.mapping.scaled_array_2d_from_array_1d(
            array_1d=g0_blurred_image_1d
        )
        g1_blurred_image_2d = lens_imaging_data_7x7.mapping.scaled_array_2d_from_array_1d(
            array_1d=g1_blurred_image_1d
        )

        assert fit.galaxy_model_image_2d_dict[g0] == pytest.approx(
            g0_blurred_image_2d, 1.0e-4
        )
        assert fit.galaxy_model_image_2d_dict[g1] == pytest.approx(
            g1_blurred_image_2d, 1.0e-4
        )
        assert fit.galaxy_model_image_2d_dict[galaxy_pix] == pytest.approx(
            inversion.reconstructed_data_2d, 1.0e-4
        )

        assert fit.model_image(return_in_2d=True) == pytest.approx(
            fit.galaxy_model_image_2d_dict[g0]
            + fit.galaxy_model_image_2d_dict[g1]
            + inversion.reconstructed_data_2d,
            1.0e-4,
        )

    def test___all_lens_fit_quantities__include_hyper_methods(
        self, lens_imaging_data_7x7
    ):

        hyper_noise_map_max = 0.2
        lens_imaging_data_7x7.hyper_noise_map_max = hyper_noise_map_max

        hyper_image_sky = al.HyperImageSky(sky_scale=1.0)

        hyper_background_noise = al.HyperBackgroundNoise(noise_scale=1.0)

        image_1d = hyper_image_sky.image_scaled_sky_from_image(
            image=lens_imaging_data_7x7._image_1d
        )

        hyper_noise_map_background_1d = hyper_background_noise.noise_map_scaled_noise_from_noise_map(
            noise_map=lens_imaging_data_7x7._noise_map_1d
        )

        galaxy_light = al.Galaxy(
            redshift=0.5,
            light_profile=al.light_profiles.EllipticalSersic(intensity=1.0),
            hyper_galaxy=al.HyperGalaxy(
                contribution_factor=1.0, noise_factor=1.0, noise_power=1.0
            ),
            hyper_model_image_1d=np.ones(9),
            hyper_galaxy_image_1d=np.ones(9),
            hyper_minimum_value=0.0,
        )

        pix = al.pixelizations.Rectangular(shape=(3, 3))
        reg = al.regularization.Constant(coefficient=1.0)
        galaxy_pix = al.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)

        tracer = al.Tracer.from_galaxies(galaxies=[galaxy_light, galaxy_pix])

        fit = al.LensImagingFit.from_lens_data_and_tracer(
            lens_data=lens_imaging_data_7x7,
            tracer=tracer,
            hyper_image_sky=hyper_image_sky,
            hyper_background_noise=hyper_background_noise,
        )

        hyper_noise_1d = tracer.hyper_noise_map_1d_from_noise_map_1d(
            noise_map_1d=lens_imaging_data_7x7._noise_map_1d
        )
        hyper_noise_map_1d = hyper_noise_map_background_1d + hyper_noise_1d
        hyper_noise_map_1d[
            hyper_noise_map_1d > hyper_noise_map_max
        ] = hyper_noise_map_max

        hyper_noise_map_2d = lens_imaging_data_7x7.mapping.scaled_array_2d_from_array_1d(
            array_1d=hyper_noise_map_1d
        )

        assert hyper_noise_map_1d == pytest.approx(fit._noise_map, 1e-4)
        assert hyper_noise_map_2d == pytest.approx(
            fit.noise_map(return_in_2d=True, return_masked=True)
        )

        blurred_profile_image_1d = tracer.blurred_profile_image_from_grid_and_convolver(
            grid=lens_imaging_data_7x7.grid,
            convolver=lens_imaging_data_7x7.convolver,
            blurring_grid=lens_imaging_data_7x7.blurring_grid,
            return_in_2d=False,
        )

        blurred_profile_image_2d = lens_imaging_data_7x7.mapping.scaled_array_2d_from_array_1d(
            array_1d=blurred_profile_image_1d
        )

        assert blurred_profile_image_1d == pytest.approx(
            fit.blurred_profile_image(return_in_2d=False), 1e-4
        )
        assert blurred_profile_image_2d == pytest.approx(
            fit.blurred_profile_image(return_in_2d=True)
        )

        profile_subtracted_image_1d = image_1d - blurred_profile_image_1d
        profile_subtracted_image_2d = lens_imaging_data_7x7.mapping.scaled_array_2d_from_array_1d(
            array_1d=profile_subtracted_image_1d
        )

        assert profile_subtracted_image_1d == pytest.approx(
            fit.profile_subtracted_image(return_in_2d=False)
        )
        assert profile_subtracted_image_2d == pytest.approx(
            fit.profile_subtracted_image(return_in_2d=True)
        )

        mapper = pix.mapper_from_grid_and_pixelization_grid(
            grid=lens_imaging_data_7x7.grid, inversion_uses_border=False
        )

        inversion = al.Inversion.from_data_1d_mapper_and_regularization(
            image_1d=profile_subtracted_image_1d,
            noise_map_1d=hyper_noise_map_1d,
            convolver=lens_imaging_data_7x7.convolver,
            mapper=mapper,
            regularization=reg,
        )

        model_image_1d = blurred_profile_image_1d + inversion.reconstructed_data_1d
        model_image_2d = lens_imaging_data_7x7.mapping.scaled_array_2d_from_array_1d(
            array_1d=model_image_1d
        )

        assert model_image_1d == pytest.approx(fit._model_data, 1e-4)
        assert model_image_2d == pytest.approx(fit.model_image(return_in_2d=True))

        residual_map_1d = af.fit_util.residual_map_from_data_mask_and_model_data(
            data=image_1d,
            mask=lens_imaging_data_7x7._mask_1d,
            model_data=model_image_1d,
        )

        residual_map_2d = lens_imaging_data_7x7.mapping.scaled_array_2d_from_array_1d(
            array_1d=residual_map_1d
        )

        assert residual_map_1d == pytest.approx(fit._residual_map, 1e-4)
        assert residual_map_2d == pytest.approx(fit.residual_map(return_in_2d=True))

        normalized_residual_map_1d = af.fit_util.normalized_residual_map_from_residual_map_noise_map_and_mask(
            residual_map=residual_map_1d,
            mask=lens_imaging_data_7x7._mask_1d,
            noise_map=hyper_noise_map_1d,
        )

        normalized_residual_map_2d = lens_imaging_data_7x7.mapping.scaled_array_2d_from_array_1d(
            array_1d=normalized_residual_map_1d
        )

        assert normalized_residual_map_1d == pytest.approx(
            fit._normalized_residual_map, 1e-4
        )
        assert normalized_residual_map_2d == pytest.approx(
            fit.normalized_residual_map(return_in_2d=True)
        )

        chi_squared_map_1d = af.fit_util.chi_squared_map_from_residual_map_noise_map_and_mask(
            residual_map=residual_map_1d,
            mask=lens_imaging_data_7x7._mask_1d,
            noise_map=hyper_noise_map_1d,
        )

        chi_squared_map_2d = lens_imaging_data_7x7.mapping.scaled_array_2d_from_array_1d(
            array_1d=chi_squared_map_1d
        )

        assert chi_squared_map_1d == pytest.approx(fit._chi_squared_map, 1e-4)
        assert chi_squared_map_2d == pytest.approx(
            fit.chi_squared_map(return_in_2d=True)
        )

        chi_squared = af.fit_util.chi_squared_from_chi_squared_map_and_mask(
            chi_squared_map=chi_squared_map_2d, mask=lens_imaging_data_7x7.mask
        )

        noise_normalization = af.fit_util.noise_normalization_from_noise_map_and_mask(
            mask=lens_imaging_data_7x7.mask, noise_map=hyper_noise_map_2d
        )

        likelihood = af.fit_util.likelihood_from_chi_squared_and_noise_normalization(
            chi_squared=chi_squared, noise_normalization=noise_normalization
        )

        assert likelihood == pytest.approx(fit.likelihood, 1e-4)

        likelihood_with_regularization = likelihood_with_regularization_from_chi_squared_regularization_term_and_noise_normalization(
            chi_squared=chi_squared,
            regularization_term=inversion.regularization_term,
            noise_normalization=noise_normalization,
        )

        assert likelihood_with_regularization == pytest.approx(
            fit.likelihood_with_regularization, 1e-4
        )

        evidence = evidence_from_inversion_terms(
            chi_squared=chi_squared,
            regularization_term=inversion.regularization_term,
            log_curvature_regularization_term=inversion.log_det_curvature_reg_matrix_term,
            log_regularization_term=inversion.log_det_regularization_matrix_term,
            noise_normalization=noise_normalization,
        )

        assert evidence == fit.evidence
        assert evidence == fit.figure_of_merit

    def test___blurred_and_model_images_of_planes_and_unmasked_blurred_profile_image_properties(
        self, lens_imaging_data_7x7
    ):
        galaxy_light = al.Galaxy(
            redshift=0.5,
            light_profile=al.light_profiles.EllipticalSersic(intensity=1.0),
        )

        pix = al.pixelizations.Rectangular(shape=(3, 3))
        reg = al.regularization.Constant(coefficient=1.0)
        galaxy_pix = al.Galaxy(redshift=1.0, pixelization=pix, regularization=reg)

        tracer = al.Tracer.from_galaxies(galaxies=[galaxy_light, galaxy_pix])

        fit = al.LensImagingFit.from_lens_data_and_tracer(
            lens_data=lens_imaging_data_7x7, tracer=tracer
        )

        blurred_profile_image_1d = tracer.blurred_profile_image_from_grid_and_convolver(
            grid=lens_imaging_data_7x7.grid,
            convolver=lens_imaging_data_7x7.convolver,
            blurring_grid=lens_imaging_data_7x7.blurring_grid,
            return_in_2d=False,
        )

        blurred_profile_image_2d = lens_imaging_data_7x7.mapping.scaled_array_2d_from_array_1d(
            array_1d=blurred_profile_image_1d
        )

        profile_subtracted_image_1d = (
            lens_imaging_data_7x7._image_1d - blurred_profile_image_1d
        )

        mapper = pix.mapper_from_grid_and_pixelization_grid(
            grid=lens_imaging_data_7x7.grid, inversion_uses_border=False
        )

        inversion = al.Inversion.from_data_1d_mapper_and_regularization(
            image_1d=profile_subtracted_image_1d,
            noise_map_1d=lens_imaging_data_7x7._noise_map_1d,
            convolver=lens_imaging_data_7x7.convolver,
            mapper=mapper,
            regularization=reg,
        )

        model_images_of_planes = fit.model_images_of_planes(return_in_2d=True)

        assert blurred_profile_image_2d == pytest.approx(
            model_images_of_planes[0], 1.0e-4
        )
        assert inversion.reconstructed_data_2d == pytest.approx(
            model_images_of_planes[1], 1.0e-4
        )



import autolens as al
import os

import numpy as np
import pytest


@pytest.fixture(name="mapper_plotter_path")
def make_mapper_plotter_setup():
    return "{}/../../../test_files/plotting/mapper/".format(
        os.path.dirname(os.path.realpath(__file__))
    )


@pytest.fixture(name="image")
def make_image():
    image = al.ScaledSquarePixelArray(array=np.ones((3, 3)), pixel_scale=1.0)
    noise_map = al.NoiseMap(array=2.0 * np.ones((3, 3)), pixel_scale=1.0)
    psf = al.PSF(array=3.0 * np.ones((3, 3)), pixel_scale=1.0)

    return al.ImagingData(image=image, pixel_scale=1.0, noise_map=noise_map, psf=psf)


@pytest.fixture(name="mask")
def make_mask():
    return al.Mask.circular(shape=((3, 3)), pixel_scale=0.1, radius_arcsec=0.1)


@pytest.fixture(name="galaxy_light")
def make_galaxy_light():
    return al.Galaxy(
        redshift=0.5, light=al.light_profiles.EllipticalSersic(intensity=1.0)
    )


@pytest.fixture(name="galaxy_mass")
def make_galaxy_mass():
    return al.Galaxy(
        redshift=0.5, mass=al.mass_profiles.SphericalIsothermal(einstein_radius=1.0)
    )


@pytest.fixture(name="grid")
def make_grid():
    return al.Grid.from_shape_pixel_scale_and_sub_size(
        shape=(100, 100), pixel_scale=0.05, sub_size=2
    )


@pytest.fixture(name="rectangular_pixelization")
def make_rectangular_pixelization():
    return al.pixelizations.Rectangular(shape=(25, 25))


@pytest.fixture(name="rectangular_mapper")
def make_rectangular_mapper(rectangular_pixelization, grid):
    return rectangular_pixelization.mapper_from_grid_and_pixelization_grid(
        grid=grid, pixelization_grid=None, inversion_uses_border=False
    )


def test__image_and_rectangular_mapper_is_output(
    image, rectangular_mapper, mapper_plotter_path, plot_patch
):
    al.mapper_plotters.plot_image_and_mapper(
        imaging_data=image,
        mapper=rectangular_mapper,
        should_plot_centres=True,
        should_plot_grid=True,
        image_pixels=[[0, 1, 2], [3]],
        source_pixels=[[1, 2], [0]],
        output_path=mapper_plotter_path,
        output_format="png",
    )
    assert mapper_plotter_path + "image_and_mapper.png" in plot_patch.paths


def test__rectangular_mapper_is_output(
    rectangular_mapper, mapper_plotter_path, plot_patch
):
    al.mapper_plotters.plot_mapper(
        mapper=rectangular_mapper,
        should_plot_centres=True,
        should_plot_grid=True,
        image_pixels=[[0, 1, 2], [3]],
        source_pixels=[[1, 2], [0]],
        output_path=mapper_plotter_path,
        output_filename="rectangular_mapper",
        output_format="png",
    )
    assert mapper_plotter_path + "rectangular_mapper.png" in plot_patch.paths



import numpy as np

import autolens as al


class MockGrid(al.Grid):
    def __new__(cls, mask, pixel_scale=1.0, sub_size=2, *args, **kwargs):
        grid = al.grid_util.grid_1d_from_mask_pixel_scales_sub_size_and_origin(
            mask=mask, pixel_scales=(pixel_scale, pixel_scale), sub_size=sub_size
        )

        obj = grid.view(cls)
        obj.mask = mask
        obj.sub_size = sub_size
        obj.sub_length = int(obj.sub_size ** 2.0)
        obj.sub_fraction = 1.0 / obj.sub_length
        obj.interpolator = None

        return obj

    def __init__(self, mask, pixel_scale=1.0, sub_size=2):
        pass


class MockBinnedGrid(al.BinnedGrid):
    pass


class MockPixelizationGrid(np.ndarray):
    def __new__(
        cls,
        arr,
        mask_1d_index_to_nearest_pixelization_1d_index=None,
        sub_mask_1d_index_to_mask_1d_index=None,
        sub_size=1,
        *args,
        **kwargs
    ):
        """A pixelization-grid of (y,x) coordinates which are used to form the pixel centres of adaptive pixelizations in the \
        *pixelizations* module.

        A *PixGrid* is ordered such pixels begin from the top-row of the mask and go rightwards and then \
        downwards. Therefore, it is a ndarray of shape [total_pix_pixels, 2]. The first element of the ndarray \
        thus corresponds to the pixelization pixel index and second element the y or x arc -econd coordinates. For example:

        - pix_grid[3,0] = the 4th unmasked pixel's y-coordinate.
        - pix_grid[6,1] = the 7th unmasked pixel's x-coordinate.

        Parameters
        -----------
        pix_grid : ndarray
            The grid of (y,x) arc-second coordinates of every image-plane pixelization grid used for adaptive source \
            -plane pixelizations.
        mask_1d_index_to_nearest_pixelization_1d_index : ndarray
            A 1D array that maps every grid pixel to its nearest pixelization-grid pixel.
        """
        obj = arr.view(cls)
        obj.mask_1d_index_to_nearest_pixelization_1d_index = (
            mask_1d_index_to_nearest_pixelization_1d_index
        )
        obj.sub_mask_1d_index_to_mask_1d_index = sub_mask_1d_index_to_mask_1d_index
        obj.sub_size = sub_size
        obj.sub_length = int(sub_size ** 2.0)
        obj.sub_fraction = 1.0 / obj.sub_length
        obj.interpolator = None
        return obj

    @property
    def mapping(self):
        return self

    def relocated_grid_from_grid(self, grid):
        return grid
